#!/usr/bin/env python3
"""
Automatic file-based configuration login based on the INI format.

The enb framework supports configuration files with `.ini` extension and format
compatible with python's configparser (https://docs.python.org/3/library/configparser.html#module-configparser),
e.g., similar to Window's INI files.

File-based configuration is used to determine the default value of enb.config.options and its CLI.
Furthermore, users may easily extend file-based configuration to their own needs.

When enb is imported, the following configuration files are read, in the given order.
Order is important because read properties overwrite any previously set values.

1. The `enb.ini` file provided with the enb library installation.

2. The `enb.ini` at the user's enb configuration dir. This path will be determined using the appdirs library,
   and will depend on the OS. In many linux boxes, this dir is `~/.config/enb`.

3. All `*.ini` files defined in the same folder as the called script, in lexicographical,
   case ignorant, order. No recursive folder search is performed.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/18"

import argparse
import os
import sys
import itertools
import glob
import ast
import configparser
import textwrap

import enb
from .. import calling_script_dir, is_enb_cli, enb_installation_dir, user_config_dir
from ..misc import Singleton as _Singleton, class_to_fqn


class AdditionalIniParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(add_help=False)
        self.add_argument("--extra_ini_paths", "--ini", nargs="+", required=False, default=[])

    def get_extra_ini_paths(self):
        extra_ini_paths = []
        parsed_options, remaining_options = self.parse_known_args()

        if is_enb_cli:
            # The --ini option is not documented in the main enb CLI for simplicity. Remove it once
            # it has been used to avoid any parsing error
            sys.argv = sys.argv[0:1] + remaining_options

        for path in parsed_options.extra_ini_paths:
            if not os.path.exists(path):
                raise SyntaxError("Input ini path {path} does not exist. Run with -h for help.")
            extra_ini_paths.append(os.path.abspath(path))
        return extra_ini_paths


class Ini(metaclass=_Singleton):
    """Class of the enb.config.ini object, that exposes file-defined configurations.
    """
    global_ini_path = os.path.join(enb_installation_dir, "config", "enb.ini")
    user_ini_path = os.path.join(user_config_dir, "enb.ini")
    local_ini_paths = sorted(glob.glob(os.path.join(calling_script_dir, "*.ini")),
                             key=lambda s: os.path.basename(s).lower())
    extra_ini_paths = AdditionalIniParser().get_extra_ini_paths()

    def __init__(self):
        super().__init__()
        # Keep track of what config files have been used to get the final result
        self.used_config_paths = []
        self.extra_ini_paths = self.extra_ini_paths if self.extra_ini_paths is not None else []

        # Parse configuration files with the specified prioritization
        self.config_parser = configparser.ConfigParser()
        self.update_from_path(self.global_ini_path)
        if os.path.exists(self.user_ini_path):
            self.update_from_path(self.user_ini_path)
        for ini_path in itertools.chain(self.local_ini_paths, self.extra_ini_paths):
            self.update_from_path(ini_path)

    def update_from_path(self, ini_path):
        """Update the current configuration by reading the contents of ini_path.
        """
        try:
            self.config_parser.read(ini_path)
            self.used_config_paths.append(ini_path)
        except configparser.ParsingError as ex:
            print(f"Found invalid ini path {ini_path} ({repr(ex).strip()}). "
                  f"Any configuration in this file will be ignored.")

    def get_key(self, section, name):
        """Return a read key value in the given section (if existing),
        after applying ast.literal_eval on the value to recognize any
        valid python literals (essentially numbers, lists, dicts, tuples, booleans and None).
        """
        try:
            # Attempt to interpret the string as a literal
            return ast.literal_eval(self.config_parser[section][name])
        except (SyntaxError, ValueError):
            # The key could not be parsed as a literal, it is returned as a string
            # (this is configparser's default)
            return self.config_parser[section][name]

    @property
    def sections_by_name(self):
        """Get a list of all configparser.Section instances, including the default section.
        """
        return dict(self.config_parser.items())

    def __repr__(self):
        s = "File-based configuration for enb, originally read in this order:\n  - "
        s += "\n  - ".join(self.used_config_paths)
        s = textwrap.indent(s, "# ")
        for section_name, section in sorted(ini.sections_by_name.items()):
            if not section:
                continue
            s += "\n\n"
            s += f"[{section_name}]\n\n"
            for k, v in sorted(section.items()):
                s += f"{k} = {v}\n"
        return s


# Export the ini object
ini = Ini()
assert ini is Ini(), "Singleton not working for enb.config.ini"


def managed_attributes(cls):
    """Decorator for classes so that their (class) attributes are set
    based on the `.ini` files found. Attributes starting with `_` are not considered.

    Values are read from the section titled as the classes fully qualified name
    (e.g., using the `[enb.aanalysis.ScalarValueAnalyzer]` header in one of the .ini files).

    Note that adding keys to that section corresponding to attributes not present
    in the definition of cls are ignored, i.e., new attributes are not added to cls.
    """
    from enb import logger

    cls_fqn = class_to_fqn(cls)

    for attribute in (k for k, v in cls.__dict__.items()
                      if not k.startswith("_")
                         and not k == "column_to_properties"
                         and not callable(v)
                         and not isinstance(v, classmethod)):
        old_value = getattr(cls, attribute)
        try:
            setattr(cls, attribute, ini.get_key(cls_fqn, attribute))
            if getattr(cls, attribute) != old_value \
                    and str(getattr(cls, attribute)) != str(old_value):
                logger.debug(f"Updating {cls_fqn}.{attribute} = {getattr(cls, attribute)} based on .ini files "
                             f"(it was {old_value})")
        except KeyError:
            found_in_bases = _add_base_attributes_recursively(attribute, cls)
            if not found_in_bases:
                logger.warn(f"The {cls_fqn} class is decorated with "
                            f"enb.ini.managed.attributes, but contains an attribute "
                            f"{repr(attribute)} which is not present in its configuration nor in any of "
                            f"its base classes. "
                            f"The class' default value in the python code definition of ({cls.__dict__[attribute]}) "
                            f"is used instead.")

    for k, v in ini.sections_by_name[cls_fqn].items():
        if not hasattr(cls, k):
            logger.warn(f"In the .ini configuration files, managed attribute {repr(k)} is defined for {cls_fqn}, "
                        f"but {cls.__name__} itself does not define that class attribute. The attribute is NOT added "
                        f"to {cls.__name__}.")

    return cls


def _add_base_attributes_recursively(attribute, base_cls, current_cls=None):
    """Given a class base_cls, attempt to get the value of attribute defined for any of its base classes,
    recursively.

    :param attribute: name of the attribute to be queried and potentially set.
    :param base_cls: class whose attribute is being managed.
    :param current_cls: used for recursion, it must be None when invoked outside this method.
    """
    if current_cls is None:
        current_cls = base_cls

    found_in_bases = False

    try:
        setattr(base_cls, attribute, ini.get_key(section=class_to_fqn(current_cls), name=attribute))
        found_in_bases = True
    except KeyError:
        for base in current_cls.__bases__:
            if base is None:
                continue
            if _add_base_attributes_recursively(attribute=attribute, base_cls=base_cls, current_cls=base):
                found_in_bases = True
                break

    return found_in_bases



#!/usr/bin/env python3
"""Utils to plot data (thinking about pyplot)
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/10"

import os
import math

import matplotlib

import enb.atable

matplotlib.use("Agg")
import matplotlib.ticker
import numpy as np
import collections
from scipy.stats import norm
from matplotlib import pyplot as plt

import enb
from enb.config import options
from enb.misc import CircularList

marker_cycle = CircularList(["o", "X", "s", "*", "p", "P", "2", "H", "X", "1", "d", "<", ">", "+"])
color_cycle = CircularList([f"C{i}" for i in list(range(4)) + list(range(6, 10)) + list(range(5, 6))])


class PlottableData:
    alpha = 0.75
    legend_column_count = 1
    color = None
    marker = None

    def __init__(self, data=None, axis_labels=None, label=None,
                 extra_kwargs=None, alpha=None, legend_column_count=None,
                 marker=None, color=None,
                 marker_size=None):
        self.data = data
        self.axis_labels = axis_labels
        self.label = label
        self.extra_kwargs = extra_kwargs if extra_kwargs is not None else {}
        self.alpha = alpha if alpha is not None else self.alpha
        self.legend_column_count = legend_column_count if legend_column_count is not None else self.legend_column_count
        self.marker = marker
        self.marker_size = marker_size
        self.color = color if color is not None else self.color

    def render(self, axes=None):
        """Render data in current figure.

        :param axes: if not None, those axes are used for plotting instead of plt
        """
        raise NotImplementedError()

    def render_axis_labels(self, axes=None):
        """Add axis labels in current figure - don't show or write the result

        :param axes: if not None, those axes are used for plotting instead of plt
        """
        raise NotImplementedError()

    def render_legend(self, axes=None):
        axes = plt if axes is None else axes
        legend = plt.legend(loc="lower center", bbox_to_anchor=(0.5, 1),

                            ncol=self.legend_column_count, edgecolor=((0, 0, 0, 0.2)))
        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor((1, 1, 1, 0))

    def __repr__(self):
        return f"{self.__class__.__name__}(color={repr(self.color)})"


class PlottableData2D(PlottableData):
    """Plot 2D data using plt.plot()
    """

    def __init__(self, x_values, y_values,
                 x_label=None, y_label=None,
                 label=None, extra_kwargs=None,
                 remove_duplicates=False,
                 alpha=None, legend_column_count=None,
                 marker=None,
                 marker_size=None):
        """
        :param x_values, y_values: values to be plotted (only a reference is kept)
        :param x_label, y_label: axis labels
        :param label: line legend label
        :param extra_kwargs: extra arguments to be passed to plt.plot
        """
        assert len(x_values) == len(y_values)
        if remove_duplicates:
            found_pairs = collections.OrderedDict()
            for x, y in zip(x_values, y_values):
                found_pairs[(x, y)] = (x, y)
            if found_pairs:
                x_values, y_values = zip(*found_pairs.values())
            else:
                x_values = []
                y_values = []

        super().__init__(data=(x_values, y_values), axis_labels=(x_label, y_label),
                         label=label, extra_kwargs=extra_kwargs, alpha=alpha,
                         legend_column_count=legend_column_count, marker_size=marker_size,
                         marker=marker)
        self.x_values = x_values
        self.y_values = y_values
        self.x_label = x_label
        self.y_label = y_label

    def render_axis_labels(self, axes=None):
        """Show the labels in label list (if not None) or in self.axis_label_list
        (if label_list None) in the current figure.

        :param axes: if not None, those axes are used for plotting instead of plt
        """
        axes = plt if axes is None else axes
        if self.x_label is not None:
            axes.set_xlabel(self.x_label)
        if self.y_label is not None:
            axes.set_ylabel(self.y_label)

    def diff(self, other, ylabel_suffix="_difference"):
        assert len(self.x_values) == len(other.x_values)
        assert len(self.y_values) == len(other.y_values)
        assert all(s == o for s, o in zip(self.x_values, other.x_values))
        return PlottableData2D(x_values=self.x_values,
                               y_values=[s - o for s, o in zip(self.y_values, other.y_values)],
                               x_label=self.x_label,
                               y_label=f"{self.y_label}{ylabel_suffix}",
                               legend_column_count=self.legend_column_count)

    def shift_y(self, constant):
        self.y_values = [y + constant for y in self.y_values]

    def shift_x(self, constant):
        self.y_values = [y + constant for y in self.y_values]


class LineData(PlottableData2D):
    def __init__(self, marker="o", marker_size=5, **kwargs):
        super().__init__(marker=marker, marker_size=marker_size, **kwargs)

    def render(self, axes=None):
        """Plot 2D data using plt.plot()

        :param axes: if not None, those axes are used for plotting instead of plt
        """
        try:
            original_kwargs = self.extra_kwargs
            extra_kwargs = dict(self.extra_kwargs)

            try:
                ms = extra_kwargs["ms"]
                del extra_kwargs["ms"]
            except KeyError:
                ms = self.marker_size

            axes.plot(self.x_values, self.y_values, label=self.label, alpha=self.alpha,
                      marker=self.marker, ms=ms,
                      **extra_kwargs)
            axes = plt if axes is None else axes
            self.render_axis_labels(axes=axes)
            if self.label is not None and self.legend_column_count != 0:
                self.render_legend(axes=axes)
        finally:
            self.extra_kwargs = original_kwargs


class ScatterData(PlottableData2D):
    def __init__(self, marker="o", alpha=0.5, marker_size=3, **kwargs):
        super().__init__(marker=marker, alpha=alpha, marker_size=marker_size, **kwargs)

    def render(self, axes=None):
        axes = plt if axes is None else axes
        self.extra_kwargs["s"] = self.marker_size

        axes.scatter(self.x_values, self.y_values, label=self.label, alpha=self.alpha,
                     marker=self.marker,
                     **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)


class BarData(PlottableData2D):
    def __init__(self, marker_size=1, **kwargs):
        super().__init__(marker_size=marker_size, **kwargs)

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.bar(self.x_values, self.y_values, label=self.label, alpha=self.alpha,
                 **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)

    def shift_y(self, constant):
        try:
            self.extra_kwargs["bottom"] += constant
        except KeyError:
            self.extra_kwargs["bottom"] = constant
        except AttributeError:
            self.extra_kwargs = dict(bottom=constant)


class StepData(PlottableData2D):
    where = "post"

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.step(self.x_values, self.y_values, label=self.label, alpha=self.alpha,
                  where=self.where, **self.extra_kwargs)
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            self.render_legend(axes=axes)


class ErrorLines(PlottableData2D):
    """One or more error lines    
    """

    def __init__(self, x_values, y_values, err_neg_values, err_pos_values, vertical=False,
                 alpha=0.5, color=None,
                 marker_size=1, cap_size=2,
                 line_width=1, **kwargs):
        """
        :param x_values, y_values: centers of the error lines
        :param err_neg_values: list of lengths for the negative part of the error
        :param err_pos_values: list of lengths for the positive part of the error
        :param vertical: determines whether the error bars are vertical or horizontal
        """
        super().__init__(x_values=x_values, y_values=y_values, remove_duplicates=False,
                         alpha=alpha, marker_size=marker_size, **kwargs)
        self.err_neg = [v if not np.isnan(v) else 0 for v in err_neg_values]
        self.err_pos = [v if not np.isnan(v) else 0 for v in err_pos_values]
        self.cap_size = cap_size if cap_size is not None else self.cap_size
        self.line_width = line_width if line_width is not None else self.line_width
        self.vertical = vertical
        self.color = color
        self.alpha = alpha
        assert len(self.x_values) == len(self.y_values)
        assert len(self.x_values) == len(self.err_pos), (
            len(self.x_values), len(self.err_pos), self.x_values, self.err_pos)

    def render(self, axes=None):
        axes = plt if axes is None else axes
        err_argument = np.concatenate((
            np.array(self.err_neg).reshape(1, len(self.err_neg)),
            np.array(self.err_pos).reshape(1, len(self.err_pos))),
            axis=0)
        # When std is computed for constant data, nan is returned. Replace those with 0 to avoid warnings.
        err_argument = err_argument[:, :len(self.x_values)]

        try:
            if self.vertical:
                axes.errorbar(self.x_values, self.y_values, yerr=err_argument,
                              fmt="-o", capsize=self.cap_size, capthick=0.5, lw=0, elinewidth=self.line_width,
                              ms=self.marker_size,
                              alpha=self.alpha,
                              **self.extra_kwargs)
            else:
                axes.errorbar(self.x_values, self.y_values, xerr=err_argument,
                              fmt="-o", capsize=self.cap_size, capthick=0.5, lw=0, elinewidth=self.line_width,
                              ms=self.marker_size,
                              alpha=self.alpha,
                              **self.extra_kwargs)
        except UserWarning as ex:
            raise ValueError(f"Error rendering {self}. x_values={repr(self.x_values)}, "
                             f"y_values={repr(self.y_values)}, self.extra_kwargs={self.extra_kwargs} "
                             f"self.err_neg={self.err_neg}, "
                             f"self.err_pos={self.err_pos},") from ex

        assert len(self.x_values) == len(self.y_values)


class HorizontalBand(PlottableData2D):
    """Plottable element that """
    alpha = 0.5
    degradation_band_count = 25
    show_bounding_lines = False

    def __init__(self, x_values, y_values, pos_height_values, neg_height_values,
                 show_bounding_lines=None,
                 degradation_band_count=None,
                 std_band_add_xmargin=False,
                 **kwargs):
        super().__init__(x_values=x_values, y_values=y_values, extra_kwargs=kwargs)
        self.extra_kwargs["lw"] = 0
        self.pos_height_values = np.array(pos_height_values)
        self.neg_height_values = np.array(neg_height_values)
        self.show_bounding_lines = show_bounding_lines if show_bounding_lines is not None else self.show_bounding_lines
        self.degradation_band_count = degradation_band_count if degradation_band_count is not None \
            else self.degradation_band_count
        self.std_band_add_xmargin = std_band_add_xmargin

    def render(self, axes=None):
        for i in range(self.degradation_band_count):
            band_fraction = i / self.degradation_band_count
            next_band_fraction = (i + 1) / self.degradation_band_count
            band_probability = 1 - (norm.cdf(band_fraction) - norm.cdf(-band_fraction))

            band_x_values = np.array(self.x_values)
            band_y_values = np.array(self.y_values)
            pos_height_values = np.array(self.pos_height_values)
            neg_height_values = np.array(self.neg_height_values)
            if self.std_band_add_xmargin:
                new_x_values = []
                new_y_values = []
                new_pos_heights = []
                new_neg_heights = []
                for x, y, pos_height, neg_height in zip(band_x_values, band_y_values, pos_height_values,
                                                        neg_height_values):
                    new_x_values.extend([x - 0.5, x + 0.5])
                    new_y_values.extend([y, y])
                    new_pos_heights.extend([pos_height, pos_height])
                    new_neg_heights.extend([neg_height, neg_height])
                band_x_values = np.array(new_x_values)
                band_y_values = np.array(new_y_values)
                pos_height_values = np.array(new_pos_heights)
                neg_height_values = np.array(new_neg_heights)

            # Fill top
            axes.fill_between(
                band_x_values,
                band_y_values + pos_height_values * band_fraction,
                band_y_values + pos_height_values * next_band_fraction,
                alpha=self.alpha * band_probability,
                color=self.color, edgecolor=None, facecolor=self.color, lw=0)
            # Fill bottom
            axes.fill_between(
                band_x_values,
                band_y_values - neg_height_values * next_band_fraction,
                band_y_values - neg_height_values * band_fraction,
                alpha=self.alpha * band_probability,
                color=self.color, edgecolor=None, facecolor=self.color, lw=0)

        if self.show_bounding_lines:
            axes.plot(self.x_values, self.y_values + pos_height_values,
                      linewidth=0.5, alpha=0.68 * self.alpha, color=self.color)
            axes.plot(self.x_values, self.y_values - neg_height_values,
                      linewidth=0.5, alpha=0.68 * self.alpha, color=self.color)

        axes = plt if axes is None else axes
        self.render_axis_labels(axes=axes)
        if self.label is not None and self.legend_column_count != 0:
            plt.legend(loc="lower center", bbox_to_anchor=(0.5, 1),
                       ncol=self.legend_column_count, facecolor=(1, 1, 1, 0))


class HorizontalLine(PlottableData):
    """Draw a horizontal line across the whole subplot.
    """

    def __init__(self, y_position, line_width=1, line_style='-', **kwargs):
        super().__init__(**kwargs)
        self.y_position = y_position
        self.line_width = line_width
        self.line_style = line_style

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.axhline(y=self.y_position, color=self.color, linestyle=self.line_style,
                     lw=self.line_width, alpha=self.alpha)


class VerticalLine(PlottableData):
    """Draw a horizontal line across the whole subplot.
    """

    def __init__(self, x_position, line_width=1, line_style='-', **kwargs):
        super().__init__(**kwargs)
        self.x_position = x_position
        self.line_width = line_width
        self.line_style = line_style

    def render(self, axes=None):
        axes = plt if axes is None else axes
        axes.axvline(x=self.x_position, color=self.color, linestyle=self.line_style,
                     lw=self.line_width, alpha=self.alpha)


@enb.parallel.parallel()
def parallel_render_plds_by_group(
        pds_by_group_name, output_plot_path, column_properties,
        global_x_label, global_y_label,
        # General figure configuration
        combine_groups=False, color_by_group_name=None, group_name_order=None,
        fig_width=None, fig_height=None,
        global_y_label_pos=None, legend_column_count=None,
        force_monochrome_group=True,
        # Axis configuration
        show_grid=None,
        semilog_y=None, semilog_y_base=10, semilog_y_min_bound=1e-10,
        group_row_margin=None,
        # Axis limits
        x_min=None, x_max=None, y_min=None, y_max=None,
        horizontal_margin=None, vertical_margin=None,
        # Optional axis labeling
        y_labels_by_group_name=None,
        x_tick_list=None, x_tick_label_list=None, x_tick_label_angle=0,
        y_tick_list=None, y_tick_label_list=None,
        plot_title=None, show_legend=True):
    """Ray wrapper for render_plds_by_group. See that method for parameter information.
    """
    try:
        if enb.parallel_ray.is_remote_node():
            os.chdir(os.path.expanduser(enb.parallel_ray.RemoteNode.remote_project_mount_path))
        else:
            os.chdir(options.project_root)

        return render_plds_by_group(pds_by_group_name=pds_by_group_name, output_plot_path=output_plot_path,
                                    column_properties=column_properties, global_x_label=global_x_label,
                                    horizontal_margin=horizontal_margin, vertical_margin=vertical_margin,
                                    y_min=y_min, y_max=y_max,
                                    force_monochrome_group=force_monochrome_group,
                                    x_min=x_min, x_max=x_max,
                                    y_labels_by_group_name=y_labels_by_group_name,
                                    color_by_group_name=color_by_group_name, global_y_label=global_y_label,
                                    combine_groups=combine_groups, semilog_y_min_bound=semilog_y_min_bound,
                                    group_row_margin=group_row_margin,
                                    group_name_order=group_name_order,
                                    fig_width=fig_width, fig_height=fig_height,
                                    global_y_label_pos=global_y_label_pos, legend_column_count=legend_column_count,
                                    show_grid=show_grid,
                                    x_tick_list=x_tick_list,
                                    x_tick_label_list=x_tick_label_list,
                                    x_tick_label_angle=x_tick_label_angle,
                                    y_tick_list=y_tick_list,
                                    y_tick_label_list=y_tick_label_list,
                                    semilog_y=semilog_y, semilog_y_base=semilog_y_base,
                                    plot_title=plot_title,
                                    show_legend=show_legend)
    except Exception as ex:
        enb.logger.error(f"Error rendering to {output_plot_path} (cwd={os.getcwd()}: {repr(ex)}")
        raise ex


def render_plds_by_group(pds_by_group_name, output_plot_path, column_properties,
                         global_x_label, global_y_label,
                         # General figure configuration
                         combine_groups=False, color_by_group_name=None, group_name_order=None,
                         fig_width=None, fig_height=None,
                         global_y_label_pos=None, legend_column_count=None,
                         force_monochrome_group=True,
                         # Axis configuration
                         show_grid=None,
                         semilog_y=None, semilog_y_base=10, semilog_y_min_bound=1e-10,
                         group_row_margin=None,
                         # Axis limits
                         x_min=None, x_max=None,
                         horizontal_margin=None, vertical_margin=None,
                         y_min=None, y_max=None,
                         # Optional axis labeling
                         y_labels_by_group_name=None,
                         x_tick_list=None, x_tick_label_list=None, x_tick_label_angle=0,
                         y_tick_list=None, y_tick_label_list=None,
                         plot_title=None,
                         show_legend=True):
    """Render lists of plotdata.PlottableData instances indexed by group name.
    Each group is rendered in a row (subplot), with a shared X axis.
    Groups can also be combined into a single row (subplot), i.e., rending all plottable data into that single subplot.
    
    When applicable, None values are substituted by
    default values given enb.config.options (guaranteed to be updated thanks to the
    @enb.parallel.parallel decorator) and the current context.
    
    Mandatory parameters:
    
    :param pds_by_group_name: dictionary of lists of PlottableData instances
    :param output_plot_path: path to the file to be created with the plot
    :param column_properties: ColumnProperties instance for the column being plotted
    :param global_x_label: x-axis label shared by all subplots (there can be just one subplot)
    :param global_y_label: y-axis label shared by all subplots (there can be just one subplot)
    
    General figure configuration:
    
    :param combine_groups: if False, each group is plotted in a different row. If True,
        all groups share the same subplot (and no group name is displayed).
    :param color_by_group_name: if not None, a dictionary of pyplot colors for the groups,
        indexed with the same keys as `pds_by_group_name`.
    :param group_name_order: if not None, it contains the order in which groups are
        displayed. If None, alphabetical, case-insensitive order is applied.
    :param fig_width: figure width. The larger the figure size, the smaller the text will look.
    :param fig_height: figure height. The larger the figure size, the smaller the text will look.
    :param global_y_label_pos: position of the global y label. Needed if the y axis has ticks with
        long labels.
    :param legend_column_count: when the legend is shown, use this many columns.
    :param force_monochrome_group: if True, all plottable data with non-None color in each group
      is set to the same color, defined by color_cycle.
    
    Axis configuration:
    
    :param show_grid: if True, or if None and options.show_grid, grid is displayed
         aligned with the major axis.
    :param semilog_y: if True, a logarithmic scale is used in the Y axis.
    :param semilog_y_base: if semilog_y is True, the logarithm base employed.
    :param semilog_y_min_bound: if semilog_y is True, make y_min the maximum of y_min and this value.
    :param group_row_margin: if provided, this margin is applied between rows of groups
    
    Axis limits:
    
    :param x_min: if not None, force plots to have this value as left end.
    :param x_max: if not None, force plots to have this value as right end.
    :param horizontal_margin: Horizontal margin to be added to the figures,
      expressed as a fraction of the horizontal dynamic range.
    :param vertical_margin: Vertical margin to be added to the figures,
      expressed as a fraction of the horizontal dynamic range.
    :param y_min: if not None, force plots to have this value as bottom end.
    :param y_max: if not None, force plots to have this value as top end.
    
    Optional axis labeling:
    
    :param y_labels_by_group_name: if not None, a dictionary of labels for the groups,
      indexed with the same keys as pds_by_group_name.
    :param x_tick_list: if not None, these ticks will be displayed in the x axis.
    :param x_tick_label_list: if not None, these labels will be displayed in the x axis.
      Only used when x_tick_list is not None.
    :param x_tick_label_angle: when label ticks are specified, they will be rotated to this angle
    :param y_tick_list: if not None, these ticks will be displayed in the y axis.
    :param y_tick_label_list: if not None, these labels will be displayed in the y axis.
      Only used when y_tick_list is not None.
    
    Global title:
    
    :param plot_title: title to be displayed.
    :param show_legend: if True, legends are added to the plot.
    """
    with enb.logger.info_context(f"Rendering {len(pds_by_group_name)} plottable data groups to {output_plot_path}",
                                 sep="...\n", msg_after=f"Done rendering into {output_plot_path}"):
        if len(pds_by_group_name) < 1:
            enb.logger.info("Warning: trying to render an empty pds_by_group_name dict. "
                            f"output_plot_path={output_plot_path}, column_properties={column_properties}. "
                            f"No analysis is performed.")
            return

        legend_column_count = options.legend_column_count if legend_column_count is None else legend_column_count
        if legend_column_count:
            for name, pds in pds_by_group_name.items():
                for pld in pds:
                    pld.legend_column_count = legend_column_count

        y_min = column_properties.hist_min if y_min is None else y_min
        y_min = max(semilog_y_min_bound, y_min if y_min is not None else 0) \
            if ((column_properties is not None and column_properties.semilog_y) or semilog_y) else y_min
        y_max = column_properties.hist_max if y_max is None else y_max

        if group_name_order is None:
            def normalize_group_label(group_name):
                if isinstance(group_name, str):
                    return group_name.strip().lower()
                else:
                    return group_name

            sorted_group_names = sorted(pds_by_group_name.keys(), key=normalize_group_label)
            if str(sorted_group_names[0]).lower() == "all":
                sorted_group_names = sorted_group_names[1:] + [str(n) for n in sorted_group_names[:1]]
        else:
            sorted_group_names = []
            for group_name in group_name_order:
                if group_name not in pds_by_group_name:
                    enb.logger.warn(
                        f"Warning: {repr(group_name)} was provided in group_name_order but is not one of the "
                        f"produced groups: {sorted(list(pds_by_group_name.keys()))}. Ignoring.")
                else:
                    sorted_group_names.append(group_name)
            for g in pds_by_group_name.keys():
                if g not in sorted_group_names:
                    if options.verbose > 2:
                        enb.logger.warn(f"Warning: {repr(g)} was not provided in group_name_order but is one of the "
                                        f"produced groups: {sorted(list(pds_by_group_name.keys()))}. Appending automatically.")
                    sorted_group_names.append(g)

        if combine_groups:
            for i, g in enumerate(sorted_group_names):
                if show_legend:
                    if (i == 0 and g.lower() != "all") or len(sorted_group_names) > 1:
                        pds_by_group_name[g][0].label = g
                for pld in pds_by_group_name[g]:
                    pld.marker = marker_cycle[i]

        y_labels_by_group_name = {g: g for g in sorted_group_names} \
            if y_labels_by_group_name is None else y_labels_by_group_name
        if color_by_group_name is None:
            color_by_group_name = {}
            for i, group_name in enumerate(sorted_group_names):
                color_by_group_name[group_name] = color_cycle[i % len(color_cycle)]
        if os.path.dirname(output_plot_path):
            os.makedirs(os.path.dirname(output_plot_path), exist_ok=True)

        fig_width = options.fig_width if fig_width is None else fig_width
        fig_height = options.fig_height if fig_height is None else fig_height
        global_y_label_pos = options.global_y_label_pos if global_y_label_pos is None else global_y_label_pos

        fig, group_axis_list = plt.subplots(
            nrows=max(len(sorted_group_names), 1) if not combine_groups else 1,
            ncols=1, sharex=True, sharey=combine_groups,
            figsize=(fig_width, max(3, 0.5 * len(sorted_group_names) if fig_height is None else fig_height)))

        if combine_groups:
            group_axis_list = [group_axis_list]
        elif len(sorted_group_names) == 1:
            group_axis_list = [group_axis_list]

        if plot_title:
            plt.suptitle(plot_title)

        semilog_x, semilog_y = False, semilog_y if semilog_y is not None else semilog_y

        if combine_groups:
            assert len(group_axis_list) == 1
            group_name_axes = zip(sorted_group_names, group_axis_list * len(sorted_group_names))
        else:
            group_name_axes = zip(sorted_group_names, group_axis_list)

        # Compute global extrema
        global_x_min = float("inf")
        global_x_max = float("-inf")
        global_y_min = float("inf")
        global_y_max = float("-inf")
        for pld in (plottable for pds in pds_by_group_name.values() for plottable in pds):
            if not isinstance(pld, PlottableData2D):
                continue
            x_values = np.array(pld.x_values, copy=False)
            if len(x_values) > 0:
                x_values = x_values[~np.isnan(x_values)]
            global_x_min = min(global_x_min, x_values.min() if len(x_values) > 0 else global_x_min)
            global_x_max = max(global_x_min, x_values.max() if len(x_values) > 0 else global_x_min)
            y_values = np.array(pld.y_values, copy=False)
            if len(y_values) > 0:
                y_values = y_values[~np.isnan(y_values)]
            global_y_min = min(global_y_min, y_values.min() if len(y_values) > 0 else global_y_min)
            global_y_max = max(global_y_min, y_values.max() if len(y_values) > 0 else global_y_min)
        if global_x_max - global_x_min > 1:
            global_x_min = math.floor(global_x_min) if not math.isinf(global_x_min) else global_x_min
            global_x_max = math.ceil(global_x_max) if not math.isinf(global_x_max) else global_x_max
        if global_y_max - global_y_min > 1:
            global_y_min = math.floor(global_y_min) if not math.isinf(global_y_min) else global_y_min
            global_y_max = math.ceil(global_y_max) if not math.isinf(global_y_max) else global_y_max
        if column_properties:
            global_x_min = column_properties.plot_min if column_properties.plot_min is not None else global_x_min
            global_x_max = column_properties.plot_max if column_properties.plot_max is not None else global_x_max

        for i, (group_name, group_axes) in enumerate(group_name_axes):
            group_color = color_by_group_name[group_name]
            for pld in pds_by_group_name[group_name]:
                pld.x_label = None
                pld.y_label = None
                d = dict()
                if force_monochrome_group:
                    pld.color = group_color if pld.color is None else pld.color
                d.update(color=pld.color)
                try:
                    pld.extra_kwargs.update(d)
                except AttributeError:
                    pld.extra_kwargs = d

                try:
                    pld.render(axes=group_axes)
                except Exception as ex:
                    raise Exception(f"Error rendering {pld} -- {group_name} -- {output_plot_path}") from ex
                semilog_x = semilog_x or (column_properties.semilog_x if column_properties else False)
                semilog_y = semilog_y or (column_properties.semilog_y if column_properties else False) or semilog_y

        for (group_name, group_axes) in zip(sorted_group_names, group_axis_list):
            if y_min != y_max:
                group_axes.set_ylim(y_min, y_max)

            if semilog_x:
                x_base = column_properties.semilog_x_base if column_properties is not None else 10
                group_axes.semilogx(base=x_base)
                group_axes.get_xaxis().set_major_locator(matplotlib.ticker.LogLocator(base=x_base))
            else:
                group_axes.get_xaxis().set_major_locator(
                    matplotlib.ticker.MaxNLocator(nbins="auto", integer=True, min_n_ticks=5))
                group_axes.get_xaxis().set_minor_locator(matplotlib.ticker.AutoMinorLocator())

            if semilog_y:
                base_y = column_properties.semilog_y_base if column_properties is not None else semilog_y_base
                group_axes.semilogy(base=base_y)
                if combine_groups or len(sorted_group_names) <= 2:
                    numticks = 11
                elif len(sorted_group_names) <= 5 and not column_properties.semilog_y:
                    numticks = 6
                elif len(sorted_group_names) <= 10:
                    numticks = 4
                else:
                    numticks = 3
                group_axes.get_yaxis().set_major_locator(matplotlib.ticker.LogLocator(base=base_y, numticks=numticks))
                group_axes.grid(True, "major", axis="y", alpha=0.2)
            else:
                group_axes.get_yaxis().set_major_locator(matplotlib.ticker.MaxNLocator(nbins="auto", integer=False))
                group_axes.get_yaxis().set_minor_locator(matplotlib.ticker.AutoMinorLocator())
            if not combine_groups:
                group_axes.get_yaxis().set_label_position("right")
                group_axes.set_ylabel(y_labels_by_group_name[group_name]
                                      if group_name in y_labels_by_group_name
                                      else enb.atable.clean_column_name(group_name),
                                      rotation=0, ha="left", va="center")

        plt.xlabel(global_x_label)
        if column_properties and column_properties.hist_label_dict is not None:
            x_tick_values = sorted(column_properties.hist_label_dict.keys())
            x_tick_labels = [column_properties.hist_label_dict[x] for x in x_tick_values]
            plt.xticks(x_tick_values, x_tick_labels)

        if global_y_label:
            fig.text(global_y_label_pos, 0.5, global_y_label, va='center', rotation='vertical')

        if options.displayed_title is not None:
            plt.suptitle(options.displayed_title)

        group_row_margin = group_row_margin if group_row_margin is not None else float(
            enb.config.options.group_row_margin)
        group_row_margin += (len(pds_by_group_name) - 6) / 24
        plt.subplots_adjust(hspace=group_row_margin)

        # Set the axis limits
        xlim = [global_x_min, global_x_max]
        ylim = [global_y_min, global_y_max]
        xlim[0] = xlim[0] if x_min is None else x_min
        xlim[1] = xlim[1] if x_max is None else x_max
        ylim[0] = ylim[0] if y_min is None else y_min
        ylim[1] = ylim[1] if y_max is None else y_max
        # Translate relative margin to absolute margin
        horizontal_margin = horizontal_margin if horizontal_margin is not None else options.horizontal_margin
        vertical_margin = vertical_margin if vertical_margin is not None else options.vertical_margin
        h_margin = horizontal_margin * (xlim[1] - xlim[0])
        v_margin = vertical_margin * (ylim[1] - ylim[0])
        xlim = [xlim[0] - h_margin, xlim[1] + h_margin]
        ylim = [ylim[0] - v_margin, ylim[1] + v_margin]
        # Apply changes to the figure
        if xlim[0] != xlim[1] and not math.isnan(xlim[0]) and not math.isnan(xlim[1]):
            plt.xlim(*xlim)
        if ylim[0] != ylim[1] and not math.isnan(ylim[0]) and not math.isnan(ylim[1]) \
                and (not semilog_y or (ylim[0] > 0 and ylim[1] > 0)):
            plt.ylim(*ylim)
            ca = plt.gca()
            for group_axes in group_axis_list:
                plt.sca(group_axes)
                plt.ylim(*ylim)
            plt.sca(ca)

        if xlim[1] < 1e-2:
            x_tick_label_angle = 90 if x_tick_label_angle is not None else x_tick_label_angle

        if x_tick_list is not None:
            if not x_tick_label_list:
                plt.xticks(x_tick_list)
            else:
                plt.xticks(x_tick_list, x_tick_label_list, rotation=x_tick_label_angle)
            plt.minorticks_off()
        if x_tick_label_list is not None:
            assert x_tick_list is not None
        if x_tick_list is None and x_tick_label_angle is not None:
            plt.xticks(rotation=x_tick_label_angle)

        for group_axes in group_axis_list:
            plt.sca(group_axes)
            if y_tick_list is not None:
                if not y_tick_label_list:
                    plt.yticks(y_tick_list)
                else:
                    plt.yticks(y_tick_list, y_tick_label_list)
                group_axes.minorticks_off()
            if y_tick_label_list is not None:
                assert y_tick_list is not None
            plt.yticks()

        show_grid = options.show_grid if show_grid is None else show_grid

        if show_grid:
            ca = plt.gca()
            plt.grid("major", alpha=0.5)
            for group_axes in group_axis_list:
                plt.sca(group_axes)
                plt.grid("major", alpha=0.5)
            plt.sca(ca)

        with enb.logger.info_context(f"Saving plot to {output_plot_path} "):
            if os.path.dirname(output_plot_path):
                os.makedirs(os.path.dirname(output_plot_path), exist_ok=True)
            plt.savefig(output_plot_path, bbox_inches="tight")
            if output_plot_path.endswith(".pdf"):
                plt.savefig(output_plot_path[:-3] + "png", bbox_inches="tight", dpi=300, transparent=True)

        plt.close()



#!/usr/bin/env python3
"""
# The config module

## Introduction

The config module deals with two main aspects:

    1. It provides the enb.config.options object with global configurations shared among enb
       modules and accessible to scripts using enb.
       These options can be accessed and set programmatically (e.g., `enb.config.options.verbose += 1`),
       and also through the CLI (see details below, or run with `-h` a python script that imports enb).

    2. It provides the enb.config.ini object to access properties defined in `.ini` files.
       These affect the default CLI values and can be easily extended by users to support
       file-based configuration. See below for details on this part.

Both `enb.config.options` and `enb.config.ini` are `argparse.Namespace` instances.
After a more detailed description of these two tools,
a summary of configuration setting priority is also provided.

## enb.config.options and CLI interface

Option configuration in enb is centralized through enb.config.options. Several key aspects should be highlighted:

    - Properties defined in `enb.config.options` are used by enb modules, and can also be used by
      scripts using enb (host scripts).

    - Many core enb functions have optional arguments with default None values.
      Those functions will often substitute None for the corresponding value in enb.config.options,
      e.g., to locate the plot output directory.

    - Scripts using enb (host scripts) may alter values in enb.config.options, e.g., before calling enb methods.
      Properties are accessed and modified with `enb.config.options.property`
      and `enb.config.property = value`, respectively. You may want to use the `from enb.config import options` line
      in your host scripts to enable less verbosity.

    - The CLI can be used to set initial values of enb.config.options properties using `-*` and `--*` arguments.
      Running with `-h` any script that imports enb will show you detailed help on all available options and
      their default values.

    - The default values for enb.config.options and its CLI is obtained through enb.config.ini, described below.


An important note should be made about the interaction between enb.config.options and ray.
When ray spawns new (local or remote) processes to serve as workers, the Options singleton
is initialized for each of those process, with the catch that ray does **not** pass the user's CLI parameters.
Therefore, different enb.config.option values would be present in the parent script and the ray workers.
To mitigate this problem, an `options` parameter is defined and passed to many these functions,
e.g., with `f.remote(options=ray.put(enb.config.options))` if f is your `@enb.parallel.parallel`-decorated function.
The `@enb.config.propagates_options` decorator provides a slightly cleaner way of automating
this mitigation.

## enb.config.ini file-based configuration

See the `enb.ini` module for further information on how this is handled.

## Effective parameter values

Based on the above description and references,
the values in `enb.config.options` will be given by the first of these options:

1. Programmatically set properties, e.g., `enb.config.options.verbose += 2`.
   The last set value is used.
2. Parameters `-*` and `-**` passed directly to the invoked script.
3. Default CLI parameters specified in any `*.ini` files in the same folder as the invoked script
   (this can be empty).
4. Default CLI parameters specified in any `*.ini` files in enb's configuration file (e.g., `~/.config/enb/enb.ini`).

From there on, many enb functions adhere to the following principle:

1. If a parameter is set to a non-None value, that value is used.
2. If a parameter with default value None is set to None or not specified,
   its value is set based on the properties in `enb.config.options`.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import datetime

# Logging tools
from .. import log
# enb.config.ini : file-based config management
from .aini import ini
# enb.config.options : CLI-based config management, defaulting to enb.config.ini
from .aoptions import propagates_options, get_options, set_options

import shutil as _shutil

options = aoptions.Options()
assert options is aoptions.Options(), "Singleton not working"


def get_banner():
    """Returns the enb banner showing the current version.
    """
    contents = f" [ Powered by enb (Experiment NoteBook) " \
               f"v{ini.get_key('enb', 'version')}" \
               f"{'::' + str(ini.get_key('enb', 'commit_hash')) if options.verbose else ''}" \
               f" ] "
    contents = f"\n{{contents:.^{_shutil.get_terminal_size()[0]}}}\n".format(contents=contents)

    return contents


def report_configuration():
    """Return a string describing the current configuration status.
    """
    return "\n".join((
        "Combined ini file configurations:",
        repr(ini),
        "",
        "Parameters of enb.config.options(after CLI parsing and potential manual changes):",
        "\n".join((f"{k} = {v}" for k, v in options.items())),
        "",
        "Logging status:",
        log.report_level_status()
    ))



import enb.plugins


class JPEGXLPlugin(enb.plugins.PluginMake):
    name = "jpegxl"
    label = "Wrapper for a reference JPEG-XL implementation"
    tags = {"data compression", "codec"}
    contrib_authors = ["See https://gitlab.com/wg1/jpeg-xl/-/blob/main/CONTRIBUTORS"]
    contrib_reference_urls = ["https://gitlab.com/wg1/jpeg-xl.git"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/jpegxl_git.zip?raw=true",
         "jpegxl_git.zip")]
    tested_on = {"linux"}



import enb.plugins


class KakaduPlugin(enb.plugins.Plugin):
    name = "kakadu"
    label = "Wrappers for Kakadu JPEG 2000"
    contrib_authors = ["Kakadu Software Pty. Ltd."]
    contrib_reference_urls = ["https://kakadusoftware.com"]
    tags = {"data compression", "codec", "privative"}
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Unit tests for the bitstream module
"""
__author__ = "Miguel Hernández-Cabronero"

import os
import unittest

import random
random.seed(0xbadc0f33)

import bitstream


class TestBitstream(unittest.TestCase):
    def test_individual_bits(self):
        # input_bits = [0,1,1,1,0,0,0,1,1,0,0,1,1,0,0,0,1]
        global_input_bits = [0, 1, 1, 1, 0, 0, 0, 1,
                             1, 0, 0, 1, 1, 0, 0, 0,
                             1]
        tmp_path = "test.bin"
        try:
            for input_bits in (global_input_bits[:i] for i in range(1, len(global_input_bits) + 1)):
                with bitstream.OutputBitStream(tmp_path) as obs:
                    for bit in input_bits:
                        obs.put_bit(bit)

                with bitstream.InputBitStream(tmp_path) as ibs:
                    read_bits = [ibs.get_bit() for _ in range(len(input_bits))]

                assert input_bits == read_bits
        finally:
            if os.path.exists(tmp_path):
                os.remove(tmp_path)

    def test_unsigned_values(self):
        tmp_path = "test.bin"
        try:
            value_count = 64
            for B in range(1, 65):
                global_input_values = [random.randint(0, 2**B - 1) for _ in range(value_count)]
                for input_values in (global_input_values[:i] for i in range(1, len(global_input_values)+1)):
                    with bitstream.OutputBitStream(tmp_path) as obs:
                        for value in input_values:
                            obs.put_unsigned_value(value=value, nbits=B)

                    with bitstream.InputBitStream(tmp_path) as ibs:
                        read_values = [ibs.get_unsigned_value(nbits=B) for _ in range(len(input_values))]

                    assert input_values == read_values
        finally:
            if os.path.exists(tmp_path):
                os.remove(tmp_path)


if __name__ == '__main__':
    unittest.main()



import enb.plugins


class FPackPlugin(enb.plugins.PluginMake):
    name = "fpack"
    label = "Wrapper for the FPACK compressor"
    tags = {"data compression", "codec"}
    contrib_authors = ["William D. Pence"]
    contrib_reference_urls = ["https://heasarc.gsfc.nasa.gov/fitsio/fpack/"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/cfitsio-3.49.tar.gz?raw=true",
         "cfitsio-3.49.tar.gz")]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Entry point for the main enb CLI.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import os
import sys
import argparse
import textwrap
import enb.plugins
from enb.config import options


def CLIParser():
    """Produce and return the main argument parser for enb.
    """
    cli_parser = argparse.ArgumentParser(
        prog="enb",
        description="CLI to the experiment notebook (enb) framework\n"
                    "(see https://github.com/miguelinux314/experiment-notebook).\n\n"
                    "Several subcommands are available; use `enb <subcommand> -h` \n"
                    "to show help about any specific command.",
        formatter_class=argparse.RawTextHelpFormatter)
    cli_parser.subparsers = cli_parser.add_subparsers(
        dest="command", required=True, description="Available enb CLI commands.")

    # plugin subcommand
    cli_parser.plugin_parser = cli_parser.subparsers.add_parser("plugin", help="Install and manage plugins.")
    cli_parser.plugin_parser.subparsers = cli_parser.plugin_parser.add_subparsers(
        description="Plugin subcommands", dest="subcommand", required=True)
    # # plugin install
    cli_parser.plugin_parser.install_parser = cli_parser.plugin_parser.subparsers.add_parser(
        "install", help="Install an available plugin.")
    cli_parser.plugin_parser.install_parser.add_argument(
        "plugin_name",
        help="Name of the plugin to be installed.")
    cli_parser.plugin_parser.install_parser.add_argument(
        "destination_dir",
        help="Path to the directory that will contain the installed plugin.")
    cli_parser.plugin_parser.install_parser.add_argument(
        # Used to trigger the desired call and save the return status
        nargs=0, dest="", action=PluginInstall)

    # # plugin list
    cli_parser.plugin_parser.list_parser = cli_parser.plugin_parser.subparsers.add_parser(
        "list", help="List available plugins.")
    cli_parser.plugin_parser.list_parser.add_argument(
        "-v", action="count", default=0, help="Show additional information about the available plugins.")
    filtering_group = cli_parser.plugin_parser.list_parser.add_argument_group("Filtering options")
    filtering_group.add_argument(
        "--exclude", nargs="*", metavar="exclude_name", default=[], type=str,
        required=False,
        help="If provided, plugins matching any of these arguments are not listed. "
             "It overwrites the filter argument(s).")
    filtering_group.add_argument(
        "filter", nargs="*",
        help="If provided, only plugins matching passed string(s) are listed.")

    cli_parser.plugin_parser.list_parser.add_argument(
        # Used to trigger the desired call and save the return status
        nargs=0, dest="", action=PluginList)

    # Help command
    cli_parser.help_parser = cli_parser.subparsers.add_parser("help", help="Show this help and exit.")

    return cli_parser


class PluginInstall(argparse.Action):
    """Action for installing an installable (Plugin, Template, etc).
    """

    def __call__(self, parser, namespace, values, option_string=None):
        plugin_name = namespace.plugin_name.strip()
        assert " " not in plugin_name, f"Plugin names cannot have spaces: {repr(plugin_name)}"
        destination_dir = namespace.destination_dir

        try:
            plugin = [p for p in enb.plugins.list_all_installables() if p.name == plugin_name][0]
        except IndexError:
            raise ValueError(
                f"Invalid plugin name {repr(plugin_name)}. Run `enb plugin list` to see available plugins.")

        if os.path.exists(destination_dir) and not issubclass(plugin, enb.plugins.Template):
            raise ValueError(f"The destination dir {repr(destination_dir)} already exists. Remove and try again.")
        try:
            plugin.install(installation_dir=destination_dir)
        except (SyntaxError, ValueError) as ex:
            enb.logger.error(f"Error installing plugin {repr(plugin_name)}: {ex}")
            sys.exit(1)

        # Set status
        setattr(namespace, self.dest, 0)


class PluginList(argparse.Action):
    """Action for listing available plugins.
    """

    def installable_matches_querys(self, installable, query_list):
        """Return true if and only if the installable matches any of the provided queries.
        """
        return any(query.lower() in installable.name.lower() or
                   (installable.label.lower() and query in installable.label.lower()) or
                   any(query.lower() in author.lower() for author in installable.contrib_authors)
                   or any(any(query.lower() in t for t in installable.tags) for f in query_list)
                   or any(any(query.lower() in t for t in installable.tested_on) for f in query_list)
                   for query in query_list)

    def __call__(self, parser, namespace, values, option_string=None):
        all_installables = enb.plugins.list_all_installables()
        filtered_installables = [
            i for i in all_installables
            if self.installable_matches_querys(i, namespace.filter if namespace.filter else [])] \
            if namespace.filter else all_installables

        if namespace.exclude:
            filtered_installables = [
                i for i in filtered_installables
                if not self.installable_matches_querys(installable=i, query_list=namespace.exclude)]

        if namespace.filter and not filtered_installables:
            print(f"No plugin matched the filter criteria ({', '.join(repr(f) for f in namespace.filter)}).")
        else:
            print(f"Showing {len(filtered_installables)} plugins", end="")
            if namespace.filter:
                print(f" matching {'any of ' if len(namespace.filter) > 1 else ''}"
                      f"{', '.join(repr(f) for f in namespace.filter)}, "
                      f"out of {len(all_installables)} available)", end="")
            else:
                print(".\nYou can add arguments to filter this list, and/or use the --exclude argument.\n"
                      "Add -v for extra information on the listed plugins", end="")
            print(".\n")

            for installable in filtered_installables:
                label = installable.label if installable.label else ''
                label = f"{label} (privative)" if "privative" in installable.tags else label
                while label[-1] == ".":
                    label = label[:-1]
                while "  " in label:
                    label = label.replace("  ", "")
                label = label.strip()
                if label:
                    label = label[0].upper() + label[1:]
                if options.verbose:
                    print("-"*20 + f"  {installable.name} :: ", end="")
                else:
                    print(f"{installable.name:>25s} :: ", end="")
                print("\n".join(textwrap.wrap(label, 100)), end="")
                print(".")
                if enb.config.options.verbose:
                    indentation_string = " " * (7 + len(" :: "))
                    if installable.authors:
                        print(textwrap.indent
                              (f"* Plugin author{'s' if len(installable.authors) != 1 else ''}:",
                               indentation_string))
                        for author in installable.authors:
                            print(textwrap.indent(f"  - {author}", indentation_string))
                    if installable.contrib_authors:
                        print(textwrap.indent(
                            f"* External software author{'s' if len(installable.contrib_authors) != 1 else ''}:",
                            indentation_string))
                        for author in installable.contrib_authors:
                            print(textwrap.indent(f"  - {author}", indentation_string))
                    if installable.contrib_reference_urls:
                        print(textwrap.indent(
                            f"* Reference URL{'s' if len(installable.contrib_reference_urls) != 1 else ''}:",
                            indentation_string))
                        for url in installable.contrib_reference_urls:
                            print(textwrap.indent(f"  - {url}", indentation_string))
                    if installable.contrib_download_url_name:
                        print(textwrap.indent(
                            f"* External software URL{'s' if len(installable.contrib_download_url_name) != 1 else ''}:",
                            indentation_string))
                        for url, name in installable.contrib_download_url_name:
                            print(textwrap.indent(f"  - {url}", indentation_string))
                            print(textwrap.indent(f"     -> <installation_dir>/{name}", indentation_string))
                    if installable.required_pip_modules:
                        print(textwrap.indent(f"* Automatically installed pip "
                                              f"{'libraries' if len(installable.required_pip_modules) > 1 else 'library'}:",
                                              indentation_string))
                        for name in installable.required_pip_modules:
                            print(textwrap.indent(f"  - {name}", indentation_string))
                    if installable.extra_requirements_message:
                        print(textwrap.indent(f"* WARNING: some requirements might need to be manually satisfied:",
                                              indentation_string))
                        print(textwrap.indent(installable.extra_requirements_message.strip(),
                                              indentation_string + "    "))
                    if installable.tags:
                        print(textwrap.indent(f"* Tag{'s' if len(installable.tags) != 1 else ''}: "
                                              f"{', '.join(repr(t) for t in installable.tags)}",
                                              indentation_string))
                    if installable.tested_on:
                        print(textwrap.indent(f"* Tested on: {', '.join(sorted(installable.tested_on))}",
                                              indentation_string))

                    print()
        print()

        print("The following plugin tags have been defined and can be used for filtering:\n")
        max_tag_length = max(len(t) for t in enb.plugins.installable.InstallableMeta.tag_to_installable.keys())
        tag_fmt_str = f"{{tag:{max_tag_length}s}}"
        for tag, installable_list in sorted(
                enb.plugins.installable.InstallableMeta.tag_to_installable.items(),
                key=lambda t: list(enb.plugins.installable.tag_to_description.keys()).index(t[0])):
            print(f"  - {tag_fmt_str.format(tag=tag)} ({len(installable_list):3d} "
                  f"{'' if len(installable_list) != 1 else ' '}plugin{'s' if len(installable_list) != 1 else ''})",
                  end="")
            try:
                print(f" {enb.plugins.installable.tag_to_description[tag]}", end="")
                print("" if enb.plugins.installable.tag_to_description[tag].endswith(".") else ".")
            except KeyError:
                print()

        print()
        print("Run"
              f"{' with -v for authorship and additional information,' if not enb.config.options.verbose else ''}"
              " with -h for full help.")

def main():
    cli_parser = CLIParser()
    options, unused_options = cli_parser.parse_known_args()

    if options.command is None:
        print("No command provided. Showing help instead.\n")
        print(f"{' [ enb help ] ':-^80s}")
        cli_parser.print_help()
    elif options.command == "help":
        cli_parser.print_help()
        return
    else:
        # Successful command run
        print()


if __name__ == '__main__':
    main()



import os
import shutil
import sys

import enb.plugins


class TestAllCodecsPlugin(enb.plugins.Template):
    name = "test-codecs"
    authors = ["Miguel Hernández-Cabronero"]
    label = "Install all codec plugins verify their availability"
    contrib_reference_urls = ["https://github.com/miguelinux314/experiment-notebook"]
    tags = {"data compression", "test"}
    tested_on = {"linux"}
    required_fields_to_help = {"include_privative_codecs": "It indicates whether privative codecs "
                                                           "(those for which codec binaries and sources are not available) "
                                                           "are included in the test.\n"
                                                           "Must be 'True' or 'False'. "}

    @classmethod
    def build(cls, installation_dir):
        super().build(installation_dir=installation_dir)
        codec_plugins = [c for c in enb.plugins.list_all_installables() if "codec" in c.tags]
        print(f"Found {len(codec_plugins)} codecs:\n\t- ")
        print("\n\t- ".join(p.name for p in codec_plugins))
        print()

        print(f"Installing {len(codec_plugins)} codecs...")
        plugin_dir = os.path.join(installation_dir, "plugins")
        os.makedirs(plugin_dir)

        for plugin in codec_plugins:
            assert cls.get_fields()["include_privative_codecs"] == "True" or cls.get_fields()[
                "include_privative_codecs"] == "False", \
                f"Invalid choice for 'include_privative_codecs': {cls.get_fields()['include_privative_codecs']}"
            if "privative" in plugin.tags and cls.get_fields()["include_privative_codecs"] == "False":
                print(f"\t... skipping privative {plugin.name}")
                continue
            print("\n" + "-"*shutil.get_terminal_size().columns + "\n")
            print(f"\t... installing {plugin.name}...")
            plugin.install(os.path.join(plugin_dir, plugin.name))
        print("All codecs installed.")
        print()

        main_script_name = "test_all_codecs.py"
        print(f"You can now run the experiment by executing the {main_script_name} script "
              f"in the installation folder, e.g., with")
        print()
        print(f"\t{os.path.basename(sys.executable)} {os.path.join(installation_dir, main_script_name)} -v")
        print()



#!/usr/bin/env python3
"""Miscellaneous tools for `enb`.

This module does not and should not import anything from enb, so that other modules
may use misc tools at definition time.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/07/11"

import builtins
import os
import sys
import csv
import re
import contextlib
import socket


def get_defining_class_name(f):
    """Return the name of the class of which f is a method, or None if not bound to any class.
    """
    try:
        return f.__qualname__.split('.<locals>', 1)[0].rsplit('.')[-2]
    except IndexError:
        return None


def remove_argparse_action(parser, action):
    """Entirely remove an action from a parser, from its subparsers and groups if it exists.
    One wonders why this is not part of the default interface...

    Adapted from https://stackoverflow.com/a/49753634.
    """
    try:
        parser._remove_action(action)
    except ValueError:
        pass
    try:
        parser._actions.remove(action)
    except ValueError:
        pass

    for g in parser._action_groups:
        try:
            g._remove_action(action)
        except ValueError:
            pass
        try:
            g._actions.remove(action)
        except ValueError:
            pass
        try:
            g._group_actions.remove(action)
        except ValueError:
            pass

    a = action
    try:
        parser._remove_action(a)
    except ValueError:
        pass
    try:
        parser._actions.remove(a)
    except ValueError:
        pass

    for o in a.option_strings:
        try:
            del parser._option_string_actions[o]
        except KeyError:
            pass

    vars_action = vars(a)
    try:
        var_group_actions = vars_action['_group_actions']
    except KeyError:
        var_group_actions = None
    if var_group_actions is not None:
        for x in var_group_actions:
            if x.dest == arg:
                var_group_actions.remove(x)


def split_camel_case(camel_string):
    """Split a camel case string like ThisIsAClass into a string like "This Is A Class".
    """
    return " ".join(re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', camel_string))


def get_all_subclasses(*base_classes):
    """Return a set of all subclasses of the classes in base_classes,
    which have been defined at this point.

    The base classes are never returned as subclasses.

    :param base_classes: the list of classes for which subclasses are to be found
    """
    base_class_set = set(base_classes)
    subclasses = set(base_classes)
    previous_length = None
    while previous_length != len(subclasses):
        previous_length = len(subclasses)
        new_codec_classes = set(c for cls in subclasses for c in cls.__subclasses__())
        subclasses.update(new_codec_classes)

    return set(cls for cls in subclasses if cls not in base_class_set)


class Singleton(type):
    """Classes using this as will only be instantiated once.
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """This method replaces the regular initializer of classes with this as their metaclass.
        *args and **kwargs are passed directly to their initializer and do not otherwise affect the Singleton behavior.
        """
        try:
            return cls._instances[cls]
        except KeyError:
            cls._instances[cls] = super().__call__(*args, **kwargs)
            return cls._instances[cls]


class ExposedProperty:
    """This method can be used to expose object properties as public callables
    that return what requesting that property would.
    """

    def __init__(self, instance, property_name):
        self.property_name = property_name
        self.instance = instance

    def __call__(self, *args, **kwargs):
        return getattr(self.instance, self.property_name)


class CircularList(list):
    """A tuned list that automatically applies modulo len(self) to the given index,
    allowing for circular, index-based access to the data (whereas itertools.cycle does
    not allow accessing elements by index).
    """

    def __getitem__(self, item):
        return super().__getitem__(item % len(self))


def class_to_fqn(cls):
    cls_fqn = f"{str(cls.__module__) + '.' if cls.__module__ is not None else ''}" \
              f"{cls.__name__}"
    return cls_fqn


def csv_to_latex_tabular(input_csv_path, output_tex_path, contains_header=True, use_booktabks=True):
    """Read a CSV table from a file and output it as a latex table to another file.
    The first row is assumed to be the header.

    :param input_csv_path: path to a file containing CSV data.
    :param output_tex_file: path where the tex contents are to be stored, ready to be `\input` in latex.
    :param contains_header: if True, the first line is assumed to be a header containing column names.
    :param use_booktabs: if True, a booktabs-based decoration style is used for the table. Otherwise,
      standard latex is used only.
    """
    with open(input_csv_path, "r") as csv_file, open(output_tex_path, "w") as tex_file:
        tex_file.write("\\begin{tabular}{")

        for i, row in enumerate(csv.reader(csv_file)):
            if i == 0:
                tex_file.write("l" * len(row) + "}\n")

            if i == 0 and contains_header:
                tex_file.write("\\toprule\n" if use_booktabks else "\\hline\n")
                tex_file.write(" & ".join(f"\\textbf{{{c}}}" for c in row).replace(
                    "_", "\\_").replace("%", "\%") + r" \\" + "\n")
                tex_file.write("\\midrule\n" if use_booktabks else "\\hline\n")
            else:
                tex_file.write(" & ".join(row).replace(
                    "_", "\\_").replace("%", "\%") + r" \\" + "\n")

        tex_file.write("\\bottomrule\n" if use_booktabks else "\\hline\n")
        tex_file.write("\\end{tabular}\n")


def get_node_ip():
    """Get the current IP address of this node.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    address = s.getsockname()[0]
    s.close()
    return address



#!/usr/bin/env python3
"""Wrappers for the reference JPEG implementation
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/04/08"

import os
import tempfile
import numpy as np
import sortedcontainers

from enb import icompression
from enb import isets
from enb import pgm
from enb import tarlite
from enb.config import options


class JPEG_LS(icompression.LosslessCodec, icompression.NearLosslessCodec, icompression.WrapperCodec):
    max_dimension_size = 65535

    def __init__(self, max_error=0, bin_dir=None, output_invocation_dir=None):
        """
        :param max_error: maximum pixelwise error allowed. Use 0 for lossless
          compression
        :param bin_dir: path to the directory that contains the
          ldc_encoder, ldc_decoder and ldc_header_tool binaries. If it is None,
          options.external_bin_base_dir is None. If this is None as well, the
          same directory of this script is used by default.
        """
        bin_dir = bin_dir if bin_dir is not None else options.external_bin_base_dir
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        assert os.path.isdir(bin_dir), f"Invalid binary dir {bin_dir}."
        jpeg_bin_path = os.path.join(bin_dir, "jpeg")

        param_dict = sortedcontainers.SortedDict()
        param_dict["ls"] = 0
        max_error = int(max_error)
        assert max_error >= 0
        param_dict["m"] = int(max_error)
        icompression.WrapperCodec.__init__(
            self, compressor_path=jpeg_bin_path, decompressor_path=jpeg_bin_path,
            param_dict=param_dict, output_invocation_dir=output_invocation_dir)

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        assert original_file_info["bytes_per_sample"] in [1, 2], \
            f"Only 1 or 2 bytes per sample, unsigned values {original_file_info['bytes_per_sample']}"
        assert original_file_info["width"] <= self.max_dimension_size
        assert original_file_info["height"] <= self.max_dimension_size
        complete_array = isets.load_array_bsq(file_or_path=original_path, image_properties_row=original_file_info)
        if original_file_info["signed"]:
            complete_array = complete_array.astype(np.int64)
            complete_array += 2 ** ((8 * original_file_info['bytes_per_sample']) - 1)
            assert (complete_array >= 0).all(), f"Error converting signed into unsigned"

        bands_per_image = min(original_file_info["component_count"],
                              self.max_dimension_size // original_file_info["height"])

        tw = tarlite.TarliteWriter()

        total_compression_time = 0

        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            stacked_size = 0
            for stack_index, start_band_index in enumerate(
                    range(0, original_file_info["component_count"], bands_per_image)):
                end_band_index_not_included = min(original_file_info["component_count"],
                                                  start_band_index + bands_per_image)
                stacked_array = np.hstack(tuple(complete_array[:, :, z].squeeze()
                                                for z in range(start_band_index, end_band_index_not_included)))
                stacked_size += stacked_array.shape[0] * stacked_array.shape[1]

                with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir,
                                                 suffix=".pgm") as tmp_stack_file:
                    pgm.write_pgm(array=stacked_array,
                                  bytes_per_sample=original_file_info["bytes_per_sample"],
                                  output_path=tmp_stack_file.name)
                    stack_compressed_path = os.path.join(tmp_dir, str(stack_index))
                    compression_results = super().compress(
                        original_path=tmp_stack_file.name,
                        compressed_path=stack_compressed_path,
                        original_file_info=original_file_info)
                    total_compression_time += compression_results.compression_time_seconds

                    tw.add_file(input_path=stack_compressed_path)
            assert stacked_size == complete_array.shape[0] * complete_array.shape[1] * complete_array.shape[2], \
                f"Total stacked size {stacked_size} does not match the expectations. " \
                f"({stacked_size} vs {complete_array.shape[0] * complete_array.shape[1] * complete_array.shape[2]}"

            tw.write(output_path=compressed_path)

            compression_results = self.compression_results_from_paths(
                original_path=original_path, compressed_path=compressed_path)
            compression_results.compression_time_seconds = total_compression_time

            return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        tr = tarlite.TarliteReader(tarlite_path=compressed_path)

        bands_per_image = min(original_file_info["component_count"],
                              self.max_dimension_size // original_file_info["height"])

        total_decompression_time = 0
        recovered_components = []
        with tempfile.TemporaryDirectory(dir=options.base_tmp_dir) as tmp_dir:
            tr.extract_all(output_dir_path=tmp_dir)

            for stack_index, start_band_index in enumerate(
                    range(0, original_file_info["component_count"], bands_per_image)):
                compressed_stack_path = os.path.join(tmp_dir, str(stack_index))

                with tempfile.NamedTemporaryFile(dir=options.base_tmp_dir, suffix=".pgm") as stack_path:
                    decompression_results = super().decompress(
                        compressed_path=compressed_stack_path, reconstructed_path=stack_path.name)
                    total_decompression_time += decompression_results.decompression_time_seconds

                    assert os.path.isfile(stack_path.name)
                    stack_array = pgm.read_pgm(input_path=stack_path.name)

                    limits = tuple(range(original_file_info["height"],
                                         stack_array.shape[1],
                                         original_file_info["height"]))

                    recovered_components.extend(np.hsplit(stack_array, limits))

        e_sum = 0
        for c in recovered_components:
            assert len(c.shape) == 2
            e_sum += c.shape[0] * c.shape[1]
        assert e_sum == original_file_info['width'] * original_file_info['height'] * original_file_info[
            'component_count'], \
            f"Wrong number of recovered pixels {e_sum}, " \
            f"expected {original_file_info['width'] * original_file_info['height'] * original_file_info['component_count']}."
        assert len(recovered_components) == original_file_info["component_count"]
        assert recovered_components[0].shape == (original_file_info["width"], original_file_info["height"])

        recovered_array = np.dstack(recovered_components)
        assert recovered_array.shape == (
            original_file_info['width'], original_file_info['height'], original_file_info['component_count'])

        if original_file_info["signed"]:
            recovered_array = recovered_array.astype(np.int64)
            recovered_array -= 2 ** ((8 * original_file_info["bytes_per_sample"]) - 1)

        isets.dump_array_bsq(array=recovered_array, file_or_path=reconstructed_path,
                             dtype=isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info))

        decompression_results = self.decompression_results_from_paths(
            compressed_path=compressed_path, reconstructed_path=reconstructed_path)
        decompression_results.decompression_time_seconds = total_decompression_time
        return decompression_results

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        s = " ".join(f"-{k} {v}" for k, v in self.param_dict.items())
        s += f" {original_path} {compressed_path}"
        return s

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return f"JPEG-LS{' PAE ' + str(self.param_dict['m']) if self.param_dict['m'] > 0 else ''}"



#!/usr/bin/env python3
"""Installable interface used for plugins and templates.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import builtins
import os
import sys
import glob
import collections
import importlib
import inspect
import shutil
import requests
import subprocess
import textwrap
import collections
import enb.misc


class InstallableMeta(type):
    """Installable classes are not meant to be instantiated, just to be defined to declare
    the presence of Installables. This metaclass is used to perform basic checks on the
    declared Installable subclasses.
    """
    installable_file_name = "__plugin__.py"
    valid_tested_on_strings = {"linux", "macos", "windows"}

    # Stores all defined Installables by name
    name_to_installable = dict()
    # Stores all defined Installables by tag
    tag_to_installable = collections.defaultdict(list)

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.name = cls.name.lower().strip() if cls.name else None
        if cls.name is not None:
            if cls.name in InstallableMeta.name_to_installable:
                raise SyntaxError(f"Installable {repr(cls)} contains a non-unique name {repr(cls.name)}.")
            else:
                InstallableMeta.name_to_installable[cls.name] = cls
        cls.label = cls.label.strip() if cls.label else cls.label
        cls.label = None if not cls.label else cls.label
        cls.authors = [a.strip() for a in cls.authors] if cls.authors else cls.authors
        cls.authors = [] if cls.authors is None else cls.authors
        cls.tags = {t.lower().strip() for t in cls.tags}
        for t in cls.tags:
            InstallableMeta.tag_to_installable[t].append(cls)

        if not all(s in cls.valid_tested_on_strings for s in cls.tested_on):
            raise SyntaxError(f"Invalid definition of {cls}: "
                              f"tested_on={repr(cls.tested_on)} contains invalid elements "
                              f"not in {repr(cls.valid_tested_on_strings)}.")

        if cls.extra_requirements_message is None and "privative" in cls.tags:
            cls.extra_requirements_message = """
            Neither source code nor binaries of this plugin can be distributed by enb. 
            It will be essentially useless unless you can provide the appropriate binaries.
            """


class Installable(metaclass=InstallableMeta):
    """Common interface to all Installable types, e.g., Plugins and Templates.
    """
    # Human friendly, expectedly unique name for the Installable
    name = None
    # Human-friendly short phrase describing the Installable.
    label = None
    # Authors of the enb Installable - by default it's the main enb maintainer.
    # Subclasses may update this as necessary.
    authors = ["Miguel Hernández-Cabronero"]
    # List of string to provide soft categorization
    tags = {}

    # Information about external ("contrib") software used by the Installable
    # Author(s) of the external software
    contrib_authors = []
    # Reference URL(s) of the external software used by the Installable
    contrib_reference_urls = []
    # List of (url, name) tuples with contents needed for retrieving and building the external software.
    # Each url must be a valid downloadable link, and name is the name set to the downloaded file
    # inside the installation dir.
    contrib_download_url_name = []

    # List of pip-installable python module names required by this Installable.
    # Subclasses must overwrite this member as necessary.
    # Can be empty if needed.
    required_pip_modules = []

    # Message shown to users when installing the Installable. It can inform about any additional
    # external software needed for this Installable to work. Typically, Installables inform about
    # apt/pacman/... requirements for the Installables to work.
    # In general, if extra_requirements_message is not None, the installation process cannot be assumed
    # to be completed automatically without user intervention.
    #
    # NOTE: the equivalent to build-essential and cmake are expected by most make-based Installables,
    # e.g., PluginMake subclasses.
    extra_requirements_message = None

    # Indicates on what platforms this Installable is known to work.
    # Can contain zero, one, or more among "linux", "darwin", "windows".
    tested_on = set()

    @classmethod
    def install(cls, installation_dir, overwrite_destination=False):
        """Install this Installable into installation_dir. By default, copy all contents
        of the Installable's source dir and install the declared pip requirements.

        :param overwrite_destination: if True, if the destination exists prior to this call,
          it is removed before installation
        """
        installation_dir = os.path.abspath(installation_dir)

        print(f"Installing {cls.name} into {installation_dir}...")

        # Warn about any manual requirements reported by the Installable
        if cls.extra_requirements_message:
            print("\tNote: This plugin contains the following message regarding additional requirements:\n")
            print(textwrap.indent(textwrap.dedent(cls.extra_requirements_message).strip(), '\t'))
            print()

        # Create output dir if needed and copy Installable contents
        if overwrite_destination and os.path.exists(installation_dir):
            try:
                shutil.rmtree(installation_dir)
            except NotADirectoryError:
                os.remove(installation_dir)
        if not os.path.exists(installation_dir):
            os.makedirs(os.path.dirname(installation_dir), exist_ok=True)

        shutil.copytree(os.path.dirname(os.path.abspath(inspect.getfile(cls))), installation_dir)

        # Install any specified pip modules - subprocess is the officially recommended way
        if cls.required_pip_modules:
            invocation = f"{sys.executable} -m pip install {' '.join(cls.required_pip_modules)}"
            print(f"Installing pip dependencies of {cls.name} with {repr(invocation)}...")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception(f"Error installing {cls.name} dependencies ({cls.required_pip_modules}). "
                                f"Status = {status} != 0.\nInput=[{invocation}].\nOutput=[{output}]")

        # Download any needed external packages before the build
        for url, name in cls.contrib_download_url_name:
            output_path = os.path.join(installation_dir, name)
            print(f"Downloading {url} into {output_path}...")
            with open(output_path, "wb") as output_file:
                output_file.write(requests.get(url, allow_redirects=True).content)

        # Custom building of the Installable
        print(f"Building {cls.name} into {installation_dir}...")
        cls.build(installation_dir=installation_dir)

    @classmethod
    def build(cls, installation_dir):
        """Method called after the main installation body, that allows further building customization
        by Installable subclasses. By default, nothing is done.

        Note that the installation dir is created before calling build.
        """
        pass

    @classmethod
    def repr(cls):
        """Get a terse representation of this Installable.
        """
        return f"{cls.__name__}(" + \
               ", ".join(f"{k}={v}" for k, v in sorted(cls.__dict__.items())
                         if not k.startswith("_")
                         and not callable(v)
                         and not inspect.ismethoddescriptor(v)) + \
               ")"

    @classmethod
    def get_help(cls):
        """Return help about this Installable.
        By default, the docstring of the selected class is returned.
        """
        return textwrap.dedent(cls.__doc__)


def import_all_installables():
    """Import all public enb Installables.

    These are recognized by containing a __plugin__.py file in
    them and a installable definition.
    """

    for plugin_path in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "**",
                                              Installable.installable_file_name),
                                 recursive=True):
        # Compute the module import name
        plugin_path = os.path.abspath(plugin_path)
        module_name = "enb.plugins"
        module_name += ".".join(plugin_path.replace(os.path.dirname(os.path.abspath(__file__)), "").split(
            os.sep)).replace(f".{Installable.installable_file_name}", f".{Installable.installable_file_name[:-3]}")

        # Importing of the __plugin__.py (installable_file_name) module
        # will fail if there is an __init__.py present which attempts to import
        # any of the external dependencies. Therefore, a more lenient version of import is used for importing the module
        original_import = builtins.__import__

        def tolerant_import(*args, **kwargs):
            try:
                return original_import(*args, **kwargs)
            except ImportError:
                print(f"Ignoring import error in {os.path.basename(os.path.dirname(plugin_path))} "
                      f"for module {args[0]}. ")

        try:
            builtins.__import__ = tolerant_import
            importlib.import_module(module_name)
        finally:
            builtins.__import__ = original_import


def list_all_installables(base_class=Installable, ignored_classes=[]):
    """Get a list of all known enb installables, sorted by name.

    :param base_class: base class used for search, e.g. Installable to search for all defined installables.
    :param ignored_classes: classes to be excluded from the returned list. In addition to these, any
      installable with name set to None is also excluded
    """
    import_all_installables()
    return sorted([cls for cls in enb.misc.get_all_subclasses(base_class)
                   if cls not in ignored_classes
                   and cls.name is not None],
                  key=lambda c: c.name.lower())


def get_installable_by_name(name):
    """Search among all defined installables for one with exactly the name provided.
    :param name: name of the installable
    :raises KeyError: if name cannot be found
    """
    for installable in list_all_installables():
        if installable.name == name:
            return installable
    raise KeyError(f"Cannot find installable {repr(name)}")


# Lean description of the intention of each tag.
tag_to_description = collections.OrderedDict()
tag_to_description["documentation"] = "Documentation examples referenced in the user manual"
tag_to_description["codec"] = "Data compression/decompression class definitions"
tag_to_description["data compression"] = "Data compression tools"
tag_to_description["template"] = "Templates formatteable into the installation dir"
tag_to_description["project"] = "Project templates, including configuration files"
tag_to_description["test"] = "Plugins for testing purposes"
tag_to_description["privative"] = "Plugins requiring additional privative software"



#!/usr/bin/env python3
"""Rull all test modules in the current working dir
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/09/19"

import os
import glob
import shutil
import unittest
import sys
import argparse
import datetime

# So that all tests can use the intended module structure transparently

parser = argparse.ArgumentParser()
parser.add_argument("-v", "--verbose", help="Be verbose? Repeat for more", action="count", default=0)
options = parser.parse_known_args()[0]

if __name__ == '__main__':
    # Clean any persistence dirs in test/
    _ = [shutil.rmtree(p)
         for p in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "*.py"))
         if os.path.isdir(p)]

    suite = unittest.TestLoader().discover(os.path.dirname(__file__))

    if options.verbose:
        print(f"Running {suite.countTestCases()} tests @ {datetime.datetime.now()}")
        print(f"{'[Params]':-^30s}")
        for param, value in options.__dict__.items():
            print(f"{param}: {value}")
        print(f"{'':-^30s}")
        print()

    unittest.TextTestRunner(verbosity=3, failfast=True).run(suite)



import enb.plugins


class CCSDS122Plugin(enb.plugins.Plugin):
    name = "ccsds122x1"
    label = "Wrappers for CCSDS 122.1"
    tags = {"data compression", "codec", "privative"}
    contrib_authors = ["Ian Blanes", "Miguel Hernández-Cabronero", "CNES, et al."]
    contrib_reference_urls = ["https://ccsds.org"]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Experiment notebook (enb) library.

Please see https://github.com/miguelinux314/experiment-notebook for further information.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/03/31"

import os as _os
import sys as _sys
import shutil as _shutil
import textwrap as _textwrap
import appdirs as _appdirs
import numpy as _np
import warnings as _warnings
import atexit as _atexit
import platform as _platform

# Make all warnings errors
_np.seterr(all="raise")
_warnings.simplefilter('error', UserWarning)

# Current installation dir of enb
enb_installation_dir = _os.path.dirname(_os.path.abspath(__file__))

# User configuration dir (e.g., ~/.config/enb in many linux distributions)
user_config_dir = _os.path.join(_os.path.abspath(_os.path.expanduser(_appdirs.user_config_dir())), f"enb")
if not _os.path.exists(user_config_dir):
    _os.makedirs(user_config_dir, exist_ok=True)
if not _os.path.isdir(user_config_dir):
    print(f"[enb.__init__.py]: user_config_dir={user_config_dir} is not a directory or could not be created.")

# Absolute, real path to the calling script's dir. Configuration files present here will overwrite
# those in `user_config_dir`.
calling_script_dir = _os.path.realpath(_os.path.dirname(_os.path.abspath(_sys.argv[0])))

# Data dir
default_base_dataset_dir = _os.path.join(calling_script_dir, "datasets")
# Persistence dir
default_persistence_dir = _os.path.join(calling_script_dir, f"persistence_{_os.path.basename(_sys.argv[0])}")

# Are we currently running the main enb CLI or the CLI for a host script? True means main enb CLI.
is_enb_cli = _os.path.basename(_sys.argv[0]) in ["__main__.py", "enb"]

# Basic tools among core modules
from . import misc
# Logging tools
from . import log
# Global configuration modules
from . import config
# Logging tools
from .log import logger
# Paralellization modules
from . import parallel
from . import parallel_ray

# Setup logging so that it is used from here on. Done here to avoid circular dependencies.
logger.selected_log_level = log.get_level(name=config.options.selected_log_level,
                                          lower_priority=config.options.verbose)
logger.show_prefixes = config.options.log_level_prefix
logger.show_prefix_level = logger.get_level(name=config.options.show_prefix_level)

if config.options.log_print and not parallel_ray.is_parallel_process():
    logger.replace_print()

# Remaining core modules
## Keystone ATable features
from . import atable
## Basic Experiment features
from . import sets
from . import experiment
## Data analysis (e.g., plotting) modules
from . import plotdata
from . import aanalysis
## Image compression modules
from . import icompression
from . import isets
from . import pgm
# Plugin and template support
from . import plugins

# Setup to be run only when enb is imported
if not parallel_ray.is_parallel_process():
    # Setup common to
    log.core(config.get_banner())

    if config.options.verbose:
        if not parallel_ray.is_ray_enabled() and _platform.system().lower() == "linux" and not config.options.no_ray:
                print(_textwrap.indent(
                    f"Ray not found. Executing with fallback parallelization engine.\n"
                    f"Note that you can install ray for faster execution and cluster support, e.g., with:\n\n\tpip install ray[default]\n\n"
                    f"You can disable this warning by passing --no_ray in the command line.\n\n",
                    prefix=" "*(_shutil.get_terminal_size().columns // 10)))
        elif parallel_ray.is_ray_enabled():
            print("Using ray for parallelization.")
        else:
            print("Using fallback pathos for parallelization.")



    __file__ = _os.path.abspath(__file__)
    if not is_enb_cli:
        _os.chdir(calling_script_dir)

    if parallel_ray.is_ray_enabled():
        _atexit.register(lambda: parallel_ray.stop_ray())
        # The list of modules loaded so far passed to any possible ray remote
        # nodes so that they don't attempt to load them again.
        config.options._initial_module_names = list(
            m.__name__ for m in _sys.modules.values() if hasattr(m, "__name__"))

# Run the setter functions on the default values too, allowing validation and normalization
config.options.update(config.options, trigger_events=True)

# Remote process invoked with ray need to have some imports added
parallel_ray.fix_imports()



#!/usr/bin/env python3
"""Unit tests for the icompression module
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/4/6"

import unittest
import tempfile
import subprocess
import glob
import os
import sys
import numpy as np
import shutil

import test_all
from enb.config import options

from enb import icompression
from enb import isets
from codec_implementations import trivial_codecs


class ConstantOutputCodec(icompression.LosslessCodec):
    """Codec that reconstructs an image with the same geometry
    but all pixels identical to a constant provided at initialization
    time
    """

    def __init__(self, reconstruct_value):
        """
        :param reconstruct_value: value that all pixels of
          the reconstructed images have.
        """
        super().__init__()
        self.reconstruct_value = reconstruct_value

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        dt = isets.iproperties_row_to_numpy_dtype(image_properties_row=original_file_info)
        a = np.full((original_file_info["width"],
                     original_file_info["height"],
                     original_file_info["component_count"]),
                    fill_value=self.reconstruct_value, dtype=dt)
        isets.dump_array_bsq(array=a, file_or_path=reconstructed_path)

    def compress(self, original_path: str, compressed_path: str, original_file_info=None):
        shutil.copyfile(original_path, compressed_path)


class TestIcompression(unittest.TestCase):

    def test_mse_pae(self):
        """Test MSE and PAE calculation for simple images
        for all supported data types.
        """
        width = 37
        height = 64
        component_count = 7
        original_constant = 10

        for signed in [True, False]:
            for bytes_per_sample in [1, 2, 4]:
                for big_endian in [True, False]:
                    if bytes_per_sample == 1 and not big_endian:
                        continue
                    dtype = ((">" if big_endian else "<") if bytes_per_sample > 1 else "") \
                            + ("i" if signed else "u") \
                            + str(bytes_per_sample)
                    sample_tag = ("s" if signed else "u") \
                                 + str(8 * bytes_per_sample) \
                                 + ("be" if big_endian else "le")

                    fill_values = [original_constant] if not signed else [-original_constant, original_constant]
                    for fill_value in fill_values:
                        array = np.full((width, height, component_count),
                                        fill_value=fill_value, dtype=dtype)
                        for error in [-3, 3]:
                            with tempfile.TemporaryDirectory() as tmp_dir, \
                                    tempfile.TemporaryDirectory() as persistence_dir, \
                                    tempfile.NamedTemporaryFile(
                                        dir=tmp_dir,
                                        suffix=f"-{component_count}x{height}x{width}_{sample_tag}.raw") as tmp_file:
                                isets.dump_array_bsq(array=array, file_or_path=tmp_file.name)

                                coc = ConstantOutputCodec(reconstruct_value=fill_value + error)
                                options.persistence_dir = persistence_dir
                                ce = icompression.LossyCompressionExperiment(
                                    codecs=[coc],
                                    dataset_paths=[tmp_file.name],
                                )
                                df = ce.get_df()
                                assert (df["pae"] == abs(error)).all(), \
                                    (df["pae"], abs(error))
                                assert (np.abs(df["mse"] - error ** 2) < (2 * sys.float_info.epsilon)).all(), \
                                    (df["mse"], abs(error))


class TestSpectralAngle(unittest.TestCase):
    def get_expected_angles_deg(self, img_a, img_b):
        """Manually obtain the vector angles in degrees"""
        width, height, component_count = img_a.shape
        assert img_a.shape == img_b.shape

        img_a = img_a.copy().astype("i4")
        img_b = img_b.copy().astype("i4")

        angles = []
        for x in range(width):
            for y in range(height):
                a = img_a[x, y, :]
                b = img_b[x, y, :]
                num = np.dot(a, b)
                den = np.sqrt(np.dot(a, a)) * np.sqrt(np.dot(b, b))

                angle = np.arccos(num / den)
                angles.append(np.rad2deg(angle))
        return angles

    def test_spectral_angle(self):
        for constant_offset in [1, 5, 10]:
            width, height, component_count = 2, 3, 4
            bytes_per_sample, signed, big_endian = 2, True, True
            row = dict(signed=signed, bytes_per_sample=bytes_per_sample, big_endian=big_endian, float=False)
            original_array = np.zeros((width, height, component_count),
                                      dtype=isets.iproperties_row_to_numpy_dtype(row))
            for x in range(width):
                for y in range(height):
                    for z in range(component_count):
                        original_array[x, y, z] = 100 * z + 10 * y + x
            reconstructed_array = original_array + constant_offset
            expected_angles = self.get_expected_angles_deg(original_array, reconstructed_array)

            tag = isets.iproperties_to_name_tag(
                width=width, height=height, component_count=component_count,
                big_endian=big_endian, bytes_per_sample=bytes_per_sample, signed=signed)

            with tempfile.TemporaryDirectory() as tmp_dir:
                with tempfile.NamedTemporaryFile(suffix="-" + tag + ".raw", dir=tmp_dir) as tmp_file:
                    isets.dump_array_bsq(original_array, tmp_file.name)
                    sa_exp = icompression.SpectralAngleTable(
                        codecs=[trivial_codecs.OffsetLossyCodec(constant_offset)],
                        dataset_paths=[tmp_file.name],
                        csv_experiment_path=os.path.join(tmp_dir, "exp_persistence.csv"),
                        csv_dataset_path=os.path.join(tmp_dir, "dataset_persistence.csv"))

                    df = sa_exp.get_df()

                    abs_diff_average_sa = abs(df.iloc[0]["mean_spectral_angle_deg"]
                                              - (sum(expected_angles) / len(expected_angles)))
                    abs_diff_max_sa = abs(df.iloc[0]["max_spectral_angle_deg"]
                                          - max(expected_angles))

                    assert abs_diff_average_sa < 1e-5, f"Wrong mean spectral angle (diff={abs_diff_average_sa})"
                    assert abs_diff_max_sa < 1e-5, f"Wrong maximum spectral angle (diff={abs_diff_max_sa})"


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Wrapper for the JPEG-XL reference implementation
"""
__author__ = "Ashwin Kumar Gururajan <ashwin.gururajan@uab.cat>"
__since__ = "2021/02/15"

import os
from enb import icompression


class JPEG_XL(icompression.LosslessCodec, icompression.PNGWrapperCodec):
    def __init__(self, quality_0_to_100=100, compression_level=7,
                 compressor_path=os.path.join(os.path.dirname(__file__), "cjxl"),
                 decompressor_path=os.path.join(os.path.dirname(__file__), "djxl")):
        """
        :param quality_0_to_100: Quality setting. Range: -inf .. 100.
        100 = mathematically lossless. Default for already-lossy input (JPEG/GIF).
        Positive quality values roughly match libjpeg quality.
        Uses jpeg_xl parameter -q and was chosen over -d maxError
        (defined by butteraugli distance) becuase of slightly better throughput
        (approx 1.2 Mp/s) under lossless mode
        :param compression_level: higher values mean slower compression
        """
        assert 3 <= compression_level <= 9
        assert 0 <= quality_0_to_100 <= 100
        icompression.PNGWrapperCodec.__init__(
            self, compressor_path=compressor_path, decompressor_path=decompressor_path,
            param_dict=dict(quality_0_to_100=quality_0_to_100, compression_level=compression_level),
            output_invocation_dir=None)

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"{original_path} {compressed_path} " \
               f"-q {self.param_dict['quality_0_to_100']} " \
               f"-s {self.param_dict['compression_level']}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"{compressed_path} {reconstructed_path}"

    @property
    def label(self):
        return f"JPEG XL"



#!/usr/bin/env python3
"""Codec wrapper for the Zstandard lossless image coder
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/07/12"

import os
import enb


class Zstandard(enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.FITSWrapperCodec):
    """Wrapper for the Zstandard codec
    All data types integer and float 16, 32, 64 can be compressed 
    """

    def __init__(self, compression_level='19', zstd_binary=os.path.join(os.path.dirname(__file__), "zstd")):
        """
        :param compression_level: 1-19, being 19 the maximum data reduction
        """
        super().__init__(compressor_path=zstd_binary,
                         decompressor_path=zstd_binary,
                         param_dict=dict(compression_level=compression_level))

    @property
    def label(self):
        return "Zstandard"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-{self.param_dict['compression_level']} -f {original_path}  -o {compressed_path}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-d -f {compressed_path} -o {reconstructed_path}"



#!/usr/bin/env python3
"""The core functionality of enb is extended by means of plugins and templates. These derive from Installable,
a class that by default copies the Installable's source contents and runs the subclass' build() method.
Python libraries available via pip can be defined for Installables, which are attempted to be satisfied before invoking
the build method.

Plugins are conceived self-contained, python modules that can assume the enb library is installed.

Templates are very similar to plugins, but use jinja to transform `.enbt` template files upon installation.

Please refer to each submodule for further information.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

from .installable import Installable, import_all_installables, list_all_installables, get_installable_by_name
from .plugin import Plugin, PluginMake
from .template import Template



#!/usr/bin/env python3
"""
Wrappers for the Kakadu codec.
For instructions on downloading and installing visit:
https://github.com/miguelinux314/experiment-notebook/blob/master/enb/pluginsplugin_kakadu/README.md
"""
__author__ = "Natalia Blasco, Ester Jara, Artur Llabrés and Miguel Hernández-Cabronero"
__since__ = "2021/08/01"

import math
import numpy as np
import os
import tempfile
from enb import icompression
from enb import isets


class Kakadu2D(icompression.WrapperCodec, icompression.LosslessCodec, icompression.LossyCodec):
    # When searching for exact PSNR values, this error is tolerated
    psnr_tolerance = 1e-3
    max_search_iterations = 32

    def __init__(self, ht=False, spatial_dwt_levels=5,
                 lossless=None, bit_rate=None, quality_factor=None, psnr=None,
                 num_threads=0):
        """Kakadu wrapper that does not apply decorrelation across bands, even if more than one is present.

            :param ht: if True, the high-throughput version of the Kakadu codec is used
            :param spatial_dwt_levels: number of spatial discrete wavelet transform levels.
              Must be between 0 and 33.
            :param lossless: if True, the compression will be lossless, and bit_rate, quality_factor or
              psnr can not be
              used. If None, a lossless compression will be performed, unless
              bit_rate/quality_factor/psnr are set,
              in which case lossless will be set to False and the compression will be lossy.
            :param bit_rate: target bits per sample for each component of the image. If bit_rate is set,
              then lossless must be None or False.
            :param quality_factor: target quality factor for a lossy compression.
              If quality_factor is set then lossless
              must be None or False. Must be between 0 and 100.
            :param psnr: target peak signal-to-noise ratio for a lossy compression.
              A binary search is performed to find which quality factor will result in the
              indicated PSNR with the tolerance given by self.psnr_tolerance.
              If psnr is set then lossless must be None or False.
            :param num_threads: number of threads with which kakadu is to be run. 0 means single thread.
        """
        assert isinstance(ht, bool), "HT must be a boolean (True/False)"
        assert spatial_dwt_levels in range(0, 34), \
            f"Invalud number of spatial DWT levels {spatial_dwt_levels}"

        lossless = lossless if lossless is not None else (bit_rate is None and quality_factor is None and psnr is None)
        if lossless:
            assert all(v is None for v in [bit_rate, quality_factor, psnr]), \
                f"Cannot set bitrate, quality factor or PSNR when lossless is requested."
        else:
            assert sum(1 for v in (bit_rate, quality_factor, psnr) if v is not None) <= 1, \
                f"Only one of bitrate, quality factor and PSNR can be set at a time."
            if bit_rate is not None:
                assert bit_rate > 0, f"Invalid bit rate {bit_rate}"
            if quality_factor is not None:
                assert 0 < quality_factor <= 100, f"Invalid quality factor {quality_factor}"
            if psnr is not None:
                assert psnr > 0, f"Invalid psnr {psnr}"
            lossless = False

        assert num_threads == int(num_threads), f"Invalid number of threads {num_threads}"
        assert num_threads >= 0, f"Invalid number of threads {num_threads}"

        icompression.WrapperCodec.__init__(
            self,
            compressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "kdu_compress"),
            decompressor_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "kdu_expand"),
            param_dict=dict(
                ht=ht,
                spatial_dwt_levels=spatial_dwt_levels,
                lossless=lossless,
                bit_rate=bit_rate,
                quality_factor=quality_factor,
                psnr=psnr,
                num_threads=num_threads))

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-i {original_path}*{original_file_info['component_count']}" \
               f"@{original_file_info['width'] * original_file_info['height'] * original_file_info['bytes_per_sample']} " \
               f"-o {compressed_path} -no_info -full -no_weights " \
               f"-num_threads {self.param_dict['num_threads']} " \
               f"Corder=LRCP " \
               f"Clevels={self.param_dict['spatial_dwt_levels']} " \
               f"Clayers={original_file_info['component_count']} " \
               f"Creversible={'yes' if self.param_dict['lossless'] else 'no'} " \
               f"Cycc=no " \
               f"Sdims=\\{{{original_file_info['width']},{original_file_info['height']}\\}} " \
               f"Nprecision={original_file_info['bytes_per_sample'] * 8} " \
               f"Sprecision={original_file_info['bytes_per_sample'] * 8} " \
               f"Nsigned={'yes' if original_file_info['signed'] else 'no'} " \
               f"Ssigned={'yes' if original_file_info['signed'] else 'no'} " \
               + (f"Qstep=0.000000001 " if self.param_dict['bit_rate'] is not None else "") \
               + f"{'Cmodes=HT' if self.param_dict['ht'] else ''} " \
                 f"{'-rate ' + str(self.param_dict['bit_rate'] * original_file_info['component_count']) if self.param_dict['bit_rate'] else ''}" \
                 f"{'Qfactor=' + str(self.param_dict['quality_factor']) if self.param_dict['quality_factor'] else ''}"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-i {compressed_path} -o {reconstructed_path} -raw_components"

    def compress(self, original_path, compressed_path, original_file_info=None):
        if self.param_dict['psnr'] is not None:
            with tempfile.NamedTemporaryFile() as tmp_file:
                br_a = 0
                br_b = 32
                iteration = 0

                psnr_error = float("inf")
                while psnr_error > self.psnr_tolerance and iteration <= self.max_search_iterations:
                    iteration += 1
                    self.param_dict['bit_rate'] = (br_b + br_a) / 2
                    compression_results = icompression.WrapperCodec.compress(
                        self, original_path, compressed_path, original_file_info=original_file_info)
                    self.decompress(
                        compressed_path,
                        reconstructed_path=tmp_file.name,
                        original_file_info=original_file_info)

                    max_error = (2 ** (8 * original_file_info["bytes_per_sample"])) - 1
                    dtype = isets.iproperties_row_to_numpy_dtype(original_file_info)
                    original_array = np.fromfile(original_path, dtype=dtype).astype(np.int64)
                    reconstructed_array = np.fromfile(tmp_file.name, dtype=dtype).astype(np.int64)
                    actual_bps = 8 * os.path.getsize(compressed_path) / original_file_info["samples"]

                    mse = np.average(((original_array - reconstructed_array) ** 2))
                    psnr = 10 * math.log10((max_error ** 2) / mse) if mse > 0 else float("inf")
                    if self.param_dict['psnr'] > psnr:
                        br_a = min(self.param_dict['bit_rate'], actual_bps)
                    else:
                        br_b = max(self.param_dict['bit_rate'], actual_bps)

                    psnr_error = abs(self.param_dict['psnr'] - psnr)

        else:
            compression_results = icompression.WrapperCodec.compress(
                self, original_path, compressed_path, original_file_info=original_file_info)

        return compression_results

    def decompress(self, compressed_path, reconstructed_path, original_file_info=None):
        temp_list = []
        temp_path = f""
        for i in range(0, original_file_info['component_count']):
            temp_list.append(tempfile.NamedTemporaryFile(suffix=".raw").name)
            if i < (original_file_info['component_count'] - 1):
                temp_path += f"{temp_list[i]},"
            else:
                temp_path += f"{temp_list[i]}"

        decompression_results = icompression.WrapperCodec.decompress(
            self, compressed_path, reconstructed_path=temp_path, original_file_info=original_file_info)

        with open(reconstructed_path, "wb") as output_file:
            for p in temp_path.split(","):
                with open(p, "rb") as component_file:
                    output_file.write(component_file.read())
        decompression_results.reconstructed_path = reconstructed_path
        return decompression_results

    @property
    def label(self):
        rate_str = "" if not self.param_dict["bit_rate"] else f" R = {self.param_dict['bit_rate']}"
        psnr_str = "" if not self.param_dict["psnr"] else f" PSNR {self.param_dict['psnr']}"
        return f"Kakadu {'HT' if self.param_dict['ht'] else ''}" \
               f"{'lossless' if self.param_dict['lossless'] else 'lossy'}{rate_str}{psnr_str}"


class KakaduMCT(Kakadu2D):
    def __init__(self, ht=False, spatial_dwt_levels=5, spectral_dwt_levels=5, lossless=None,
                 bit_rate=None, quality_factor=None, psnr=None):
        """
        :param spectral_dwt_levels: number of spectral discrete wavelet transform levels. Must be between 0 and 32.
        """
        assert 0 <= spectral_dwt_levels <= 32, f"Invalid number of spectral levels"
        Kakadu2D.__init__(self, ht=ht, spatial_dwt_levels=spatial_dwt_levels, lossless=lossless,
                          bit_rate=bit_rate, quality_factor=quality_factor, psnr=psnr)
        self.param_dict["spectral_dwt_levels"] = spectral_dwt_levels

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return Kakadu2D.get_compression_params(
            self,
            original_path=original_path,
            compressed_path=compressed_path,
            original_file_info=original_file_info) + \
               f" Mcomponents={original_file_info['component_count']} " \
               f"Mstage_inputs:I1=\\{{0,{original_file_info['component_count'] - 1}\\}} " \
               f"Mstage_outputs:I1=\\{{0,{original_file_info['component_count'] - 1}\\}} " \
               f"Mstage_collections:I1=\\{{{original_file_info['component_count']},{original_file_info['component_count']}\\}} " \
               f"Mstage_xforms:I1=\\{{DWT," \
               + ('1' if self.param_dict['lossless'] else '0') + \
               f",4,0,{self.param_dict['spectral_dwt_levels']}\\}} " \
               f"Mvector_size:I4={original_file_info['component_count']} " \
               f"Mvector_coeffs:I4=0 Mnum_stages=1 Mstages=1"

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f"-i {compressed_path} -o {reconstructed_path} -raw_components "

    @property
    def label(self):
        return super().label.replace("Kakadu", "Kakadu MCT")



import enb.plugins


class HEVCPlugin(enb.plugins.PluginMake):
    name = "hevc"
    label = "Wrapper for the HEVC / H.265 codec"
    tags = {"data compression", "codec"}
    contrib_authors = ["ISO", "ITU", "IEC"]
    contrib_reference_urls = ["https://hevc.hhi.fraunhofer.de/"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/HM-master.zip?raw=true",
         "HM-master.zip")]
    extra_requirements_message = """
    The cmake tool is needed to build this plugin, if not present already. 
    You can install cmake as follows: 
    
    * Ubuntu: `sudo apt install cmake`
    * MacOS:
        1. Install homebrew from http://brew.sh
        2. `brew install gcc@9`
    """
    tested_on = {"linux"}



import os
import importlib
import glob

# By default, all modules under this folder are imported
for plugin_dir in (os.path.abspath(path)
                   for path in glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "*"))
                   if os.path.isdir(path) and os.path.isfile(os.path.join(path, f"__init__.py"))):
    importlib.import_module(f"{os.path.basename(os.path.dirname(os.path.abspath(__file__)))}"
                            f".{os.path.basename(plugin_dir)}")



#!/usr/bin/env python3
"""Evaluate the availability of data compressor/decompressor pairs (codecs) defined for enb.

By default, all detected, non-"abstract" codecs are tested.
If one or more command line arguments are passed, only codec classes matching that name will be tested.
Empty codec selections are invalid.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2021/07/06"

import sys
import os
import tempfile
import filecmp
import glob
import re
import traceback
import importlib

import enb
from enb.config import options


def log_event(s):
    if options.verbose:
        s = f" {s}..."
        print(f"\n{s:->100s}\n")


class AvailabilityExperiment(enb.experiment.Experiment):
    def __init__(self, codecs):
        super().__init__(tasks=codecs, dataset_info_table=enb.isets.ImagePropertiesTable)

    @enb.atable.column_function([
        enb.atable.ColumnProperties("is_working", label="Does the codec work for this input?"),
        enb.atable.ColumnProperties("is_lossless", label="Was compression lossless?"),
        enb.atable.ColumnProperties("cr_dr", label="Compression ratio respect to the dynamic range."),
        enb.atable.ColumnProperties("error_str", label="Error string for this execution"),
    ])
    def set_availability_columns(self, index, row):
        file_path, codec = self.index_to_path_task(index)
        try:
            with tempfile.NamedTemporaryFile() as tmp_compressed_file, \
                    tempfile.NamedTemporaryFile() as tmp_reconstructed_file:
                os.remove(tmp_compressed_file.name)
                os.remove(tmp_reconstructed_file.name)
                codec.compress(
                    original_path=file_path,
                    compressed_path=tmp_compressed_file.name,
                    original_file_info=self.get_dataset_info_row(file_path))
                if not os.path.isfile(tmp_compressed_file.name) or not os.path.getsize(tmp_compressed_file.name):
                    raise enb.icompression.CompressionException(
                        f"[E]rror compressing {index} -- did not produce a file")
                codec.decompress(
                    compressed_path=tmp_compressed_file.name,
                    reconstructed_path=tmp_reconstructed_file.name,
                    original_file_info=self.get_dataset_info_row(file_path))
                row["is_working"] = True
                row["is_lossless"] = filecmp.cmp(file_path, tmp_reconstructed_file.name)
                row["cr_dr"] = self.get_dataset_info_row(file_path)["samples"] \
                               * self.get_dataset_info_row(file_path)["dynamic_range_bits"] \
                               / (8 * os.path.getsize(tmp_compressed_file.name))
                row["error_str"] = "No error"
        except Exception as ex:
            row["error_str"] = repr(traceback.format_exc())
            row["is_working"] = False
            row["is_lossless"] = False
            row["cr_dr"] = 0


class CodecSummaryTable(enb.atable.SummaryTable):
    # These are the availability levels supported so far
    UNAVAILABLE, NOT_LOSSLESS, LOSSLESS = range(3)
    availability_to_label = {
        UNAVAILABLE: "Unavailable",
        NOT_LOSSLESS: "Not lossless",
        LOSSLESS: "Lossless",
    }
    availability_modes = [UNAVAILABLE, NOT_LOSSLESS, LOSSLESS]

    def split_groups(self, reference_df=None, include_all_group=None):
        """Each row of the table corresponds to a single codec (label)."""
        reference_df = reference_df if reference_df is not None else self.reference_df
        return reference_df.groupby("task_label")

    @enb.atable.column_function("type_to_availability", label="\nCodec availability for different data types",
                                has_dict_values=True,
                                plot_min=min(availability_modes) - 0.2,
                                plot_max=max(availability_modes) + 0.2)
    def set_type_bands_to_availability(self, index, row):
        local_df = self.label_to_df[index]
        type_to_availability = dict()
        for (type_name, component_count), type_df in local_df.groupby(["type_name", "component_count"]):
            if type_name[1] == "8":
                type_name = f"{type_name[:1]} 8{type_name[2:]}"
            type_name = f"{type_name[:3]} bit {type_name[3:]}"
            if type_name[0] == "f":
                type_name = f"Float {type_name[1:]}"
            elif type_name[0] == "u":
                type_name = f"Unsigned int {type_name[1:]}"
            elif type_name[0] == "s":
                type_name = f"Signed int {type_name[1:]}"

            if type_name[-2:] == "be":
                type_name = f"{type_name[:-2]}   big endian"
            elif type_name[-2:] == "le":
                type_name = f"{type_name[:-2]} little endian"

            key = f"{type_name} {component_count:02d} bands"
            if not type_df["is_working"].all():
                type_to_availability[key] = CodecSummaryTable.UNAVAILABLE
            elif type_df["is_lossless"].all():
                type_to_availability[key] = CodecSummaryTable.LOSSLESS
            else:
                type_to_availability[key] = CodecSummaryTable.NOT_LOSSLESS
        row[_column_name] = type_to_availability


if __name__ == '__main__':
    if options.verbose:
        print(f"{' [ Codec Availability Test Script ] ':=^100s}")

    # Plugin import -- these are needed so that codecs get defined and can
    # be automatically retrieved below (ignore IDE non-usage warnings).
    # NOTE: the build script should have created the `plugins` folder with all these codecs

    log_event("Importing plugin modules")
    plugin_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "plugins")
    assert os.path.isdir(plugin_dir), \
        "The build script should have placed everything under plugins, " \
        "please report this if you feel it is a bug."
    for d in [p for p in glob.glob(os.path.join(plugin_dir, "*")) if os.path.exists(os.path.join(p, "__init__.py"))]:
        print(f"Importing {os.path.basename(d)}...")
        importlib.import_module(f"{os.path.basename(plugin_dir)}.{os.path.basename(d)}")

    # Make sure data are ready
    log_event("Preparing test dataset")
    from generate_test_images import generate_test_images

    generate_test_images()
    options.base_dataset_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

    # This part searches for all defined codecs so far. Import new codecs to make them appear in the list
    # It also filters away any class with 'abstract' in the name.
    codec_classes = enb.misc.get_all_subclasses(
        enb.icompression.LosslessCodec, enb.icompression.NearLosslessCodec, enb.icompression.LossyCodec)

    # Remove any unwanted classes from the analysis
    codec_classes = set(c for c in codec_classes if "abstract" not in c.__name__.lower())
    codec_classes = set(c for c in codec_classes if "fapec" not in c.__name__.lower())
    codec_classes = set(c for c in codec_classes if "magli" not in c.__name__.lower())
    # Specialized lcnl configurations
    codec_classes = set(c for c in codec_classes if "greenbook" not in c.__name__.lower())
    codec_classes = set(c for c in codec_classes if ("mhdc" not in c.__name__.lower()
                                                     or ("mhdc" in c.__name__.lower()
                                                         and c.__name__ == "MHDC_POT")))

    filter_args = [a for a in sys.argv[1:] if not a.startswith("-")]
    if filter_args:
        log_event(f"Filtering for {', '.join(repr(s) for s in filter_args)}")
        codec_classes = [c for c in codec_classes
                         if any(s.strip().lower() in c.__name__.strip().lower() for s in filter_args)]
        if not codec_classes:
            log_event("Error: no codec matched the filter strings "
                      f"({', '.join(repr(v) for v in sys.argv[1:] if not v.startswith('-'))}). "
                      f"Exiting now.")
            raise ValueError(sys.argv)

    if options.verbose:
        log_event(f"The following {len(codec_classes)} codecs have been found"
                  f"{' (after filtering)' if len(sys.argv) > 1 else ''}")
        for c in sorted(codec_classes, key=lambda cls: cls.__name__.lower()):
            print(f"\t:: {c.__name__}")

    # Run the experiment
    log_event(f"Running the experiment. This might take some time...")
    exp = AvailabilityExperiment(codecs=sorted((cls() for cls in codec_classes), key=lambda codec: codec.label))
    full_availability_df = exp.get_df()

    summary_df = CodecSummaryTable(
        csv_support_path=os.path.join(options.persistence_dir, f"persistence_summary.csv"),
        full_df=full_availability_df).get_df()
    analyzer = enb.aanalysis.DictNumericAnalyzer()

    all_keys = sorted(list(summary_df.iloc[0]["type_to_availability"].keys()))


    selected_keys = all_keys
    key_to_x = {k: i for i, k in enumerate(selected_keys)}
    for i in range(4, len(selected_keys), 4):
        for k in all_keys[i:]:
            key_to_x[k] += 1.75

    analyzer.get_df(
        full_df=summary_df,
        target_columns=["type_to_availability"],
        group_by="group_label",
        key_to_x=key_to_x,
        x_tick_list=[key_to_x[k] for k in selected_keys],
        x_tick_label_list=selected_keys,
        x_tick_label_angle=90,
        fig_width=10,
        fig_height=15,
        y_tick_list=CodecSummaryTable.availability_modes,
        y_tick_label_list=[CodecSummaryTable.availability_to_label[m] for m in
                           CodecSummaryTable.availability_modes],
        group_row_margin=-0.25,
        y_min=-0.3,
        y_max=2.3,
        global_x_label="Data type",
        show_count=False,
        global_y_label="Availability"
    )

    log_event("Test all codecs successfully completed")
    print(f"You can now see the availability plots in .pdf and .png format at {options.plot_dir}")



import enb.plugins


class FPZIPPlugin(enb.plugins.PluginMake):
    name = "fpzip"
    label = "Wrappers for FPZIP codecs"
    tags = {"data compression", "codec"}
    contrib_authors = ["Peter Lindstrom", "Martin Isenburg"]
    contrib_reference_urls = ["https://doi.org/10.1109/TVCG.2006.143"]
    contrib_download_url_name = [
        ("https://github.com/miguelinux314/experiment-notebook/blob/dev/contrib/fpzip-1.3.0.tar.gz?raw=true",
         "fpzip-1.3.0.tar.gz")]
    tested_on = {"linux"}



#!/usr/bin/env python3
"""Tools to execute functions in parallel_decorator using the ray library
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2019/11/21"

import logging
import shutil
import string
import math
import threading
import time
import sys
import os
import builtins
import subprocess
import random
import pandas as pd
import psutil
import signal
import glob
import platform
import ast
import importlib
import textwrap

import enb
from . import config
from .config import options
from . import log
from .log import logger

# Ray is only expected to be available on linux nodes when clustering is desired.
try:
    import ray

    _ray_present = True
except ImportError as ex:
    _ray_present = False

_ssh_present = shutil.which("ssh") is not None
_sshfs_present = shutil.which("sshfs") is not None
_dpipe_present = shutil.which("dpipe") is not None
_ray_cli_present = shutil.which("ray") is not None


def is_ray_enabled():
    """Return True if and only if ray is available and the current platform is one
    of the supported for ray clustering (currently only linux).
    """
    if not _ray_present:
        return False
    elif platform.system().lower() == "windows":
        return False
    elif not _ray_present:
        logger.debug("The ray library is available but the ray command is not available. Please fix your path.")
        return False
    elif options.no_ray:
        return False
    return True


class HeadNode:
    """Class used to initialize and stop a ray head node.

    The stop() method must be called after start(),
    or a ray cluster will remain active.
    """

    def __init__(self, ray_port, ray_port_count):
        if not is_ray_enabled():
            raise RuntimeError("The ray module is not present or is not available. "
                               "You can install it with 'pip install ray'.")

        assert ray_port == int(ray_port), ray_port
        assert ray_port_count == int(ray_port_count), ray_port_count
        assert 1025 <= ray_port, ray_port
        assert ray_port_count >= 1
        assert ray_port + ray_port_count - 1 <= 65535, (ray_port, ray_port_count, ray_port + ray_port_count - 1)
        self.ray_port = int(ray_port)
        self.ray_port_count = int(ray_port_count)
        self.session_password = ''.join(random.choices(string.ascii_letters, k=128))
        # List of RemoteNode instances started by this head node
        self.remote_nodes = []
        self.address = self.get_node_ip()

    def start(self):
        with logger.info_context(f"Stoping any previous instance of ray..."):
            invocation = "ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise Exception("Status = {} != 0.\nInput=[{}].\nOutput=[{}]".format(
                    status, invocation, output))

        with logger.info_context(f"Starting ray on port {self.ray_port}"):
            invocation = f"ray start --head " \
                         f"--include-dashboard false " \
                         f"--port {self.ray_port} " \
                         f"--ray-client-server-port {self.ray_port + 1} " \
                         f"--node-manager-port {self.ray_port + 2} " \
                         f"--object-manager-port {self.ray_port + 3} " \
                         f"--gcs-server-port  {self.ray_port + 4} " \
                         f"--min-worker-port  {self.ray_port + 5} " \
                         f"--max-worker-port  {self.ray_port + self.ray_port_count - 1} " \
                         f"--redis-password='{self.session_password}' " \
                         + (f" --num-cpus {options.ray_cpu_limit}" if options.ray_cpu_limit else "")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error starting head ray process\n"
                                   f"Command: {repr(invocation)}. Ouput:\n{output}")

        with logger.info_context(f"Initializing ray client on local port {self.ray_port}"):
            # The exported environment consists of all *.py files.
            # Furthermore, the list of current modules minus the ones found after
            # initializing enb is passed as an environment variable to
            # allow the needed imports before parallel_decorator methods are invoked.
            excludes = [os.path.relpath(p, options.project_root)
                        for p in glob.glob(os.path.join(options.project_root, "**", "*"), recursive=True)
                        if os.path.isfile(p) and not p.endswith(".py")]

            modules_needed_remotely = [m.__name__ for m in sys.modules.values()
                                       if hasattr(m, "__name__") and m.__name__ not in options._initial_module_names
                                       and not m.__name__.startswith("_")]

            ray.init(address=f"localhost:{self.ray_port}",
                     _redis_password=self.session_password,
                     runtime_env=dict(working_dir=options.project_root,
                                      excludes=excludes,
                                      env_vars=dict(_needed_modules=str(modules_needed_remotely))),
                     # runtime_env=dict(py_modules=py_modules),
                     logging_level=logging.CRITICAL if options.verbose <= 1 else logging.INFO)

        if options.ssh_cluster_csv_path:
            failing_tool = None
            if not _ssh_present:
                failing_tool = "ssh"
            elif not _sshfs_present:
                failing_tool = "sshfs"
            elif not _dpipe_present:
                failing_tool = "dpipe"
            if failing_tool:
                enb.logger.warn(f"An enb cluster configuration was selected ({repr(options.ssh_cluster_csv_path)}) "
                                f"but {failing_tool} was not found in the path. "
                                f"No remote nodes will be used in this session.\n"
                                f"Please install {failing_tool} in your system and/or fix the path it and retry.")
            else:
                if not os.path.exists(options.ssh_cluster_csv_path):
                    raise ValueError(f"The cluster configuration file was set to {repr(options.ssh_cluster_csv_path)}"
                                     f"but it does not exist. "
                                     f"Either set enb.config.options.ssh_cluster_csv_path to None "
                                     f"or to an existing file. See "
                                     f"https://miguelinux314.github.io/experiment-notebook/installation.html "
                                     f"for more details.")

                self.remote_nodes = self.parse_cluster_config_csv(options.ssh_cluster_csv_path)
                with logger.info_context(f"Connecting {len(self.remote_nodes)} remote nodes.",
                                         msg_after=f"Done connecting {len(self.remote_nodes)} remote nodes."):
                    connected_nodes = []
                    for rn in self.remote_nodes:
                        with logger.info_context(f"Connecting to {rn}"):
                            try:
                                rn.connect()
                                connected_nodes.append(rn)
                            except RuntimeError as ex:
                                print(
                                    f"Error connecting to {rn}: {repr(ex)}. Execution will continue without this node.")
                    self.remote_nodes = connected_nodes

                logger.info(f"All ({len(self.remote_nodes)}) nodes connected")

    def stop(self):
        with logger.info_context("Stopping ray server."):
            # This tiny delay allows error messages from child processes to reach the
            # orchestrating process for logging.
            # It might need to be tuned for distributed computation across networks.
            time.sleep(options.preshutdown_wait_seconds)

            invocation = "ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                logger.error("Error stopping ray process. You might need to run `ray stop` manually.\n"
                             f"Command: {repr(invocation)}. Ouput:\n{output}")

        with logger.info_context("Disconnecting from ray"):
            ray.shutdown()

        if self.remote_nodes:
            with logger.info_context("Stopping remote nodes...\n", msg_after=f"disconnected all remote nodes."):
                for rn in self.remote_nodes:
                    rn.disconnect()
            self.remote_nodes = []

    def parse_cluster_config_csv(self, csv_path):
        """Read a CSV defining remote nodes and return a list with as many RemoteNode as
        data rows in the CSV.
        """

        def clean_value(s, default_value=None):
            return s if (isinstance(s, str) and s) or not math.isnan(s) else default_value

        return [RemoteNode(
            address=clean_value(row["address"], None),
            ssh_user=clean_value(row["ssh_user"], os.getlogin()),
            ssh_port=clean_value(row["ssh_port"], 22),
            local_ssh_file=clean_value(row["local_ssh_file"], None),
            cpu_limit=clean_value(row["cpu_limit"], None),
            head_node=self)
            for _, row in pd.read_csv(csv_path).iterrows()]

    def get_node_ip(self):
        """Adapted from https://stackoverflow.com/a/166589/992926.
        """
        assert not is_parallel_process()

        try:
            return self._head_node_address
        except AttributeError:
            self._head_node_address = enb.misc.get_node_ip()
            return self._head_node_address

    @property
    def status_str(self):
        """Return a string reporting the status of the cluster"""
        return f"The current enb/ray cluster consists of:\n" \
               f"\t- {len(self.remote_nodes) + 1} total nodes.\n" + \
               ((f"\t- {len(self.remote_nodes)} remote nodes:\n\t\t * " +
                 f'\n\t\t * '.join(str(rn) for rn in self.remote_nodes) + "\n") if self.remote_nodes else "") + \
               f"\t- {int(ray.cluster_resources()['CPU'])} virtual CPU cores.\n" \
               f"\t- {int(ray.cluster_resources()['GPU']) if 'GPU' in ray.cluster_resources() else 0} GPU devices."


# Single HeadNode instance that controls the ray cluster
_head_node = None


class RemoteNode:
    """Represent a remote node of the cluster, with tools to connect via ssh.
    """
    remote_project_mount_path = "~/.enb_remote"

    def __init__(self, address, ssh_port, head_node, ssh_user=None, local_ssh_file=None, cpu_limit=None):
        assert is_ray_enabled()

        self.address = address
        self.ssh_user = ssh_user
        self.ssh_port = ssh_port
        self.local_ssh_file = local_ssh_file
        self.head_node = head_node
        self.mount_popen = None
        self.cpu_limit = cpu_limit
        if self.cpu_limit is not None and self.cpu_limit <= 0:
            self.cpu_limit = None

    def connect(self):
        assert not is_parallel_process()

        # Create remote_node_folder_path on the remote host if not existing
        with logger.info_context(f"Stopping ray on {self.address}"):
            invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                         f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                         f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error stopping remote ray process on {self}.\n"
                                   f"Command: {repr(invocation)}. Ouput:\n{output}")

        # Create remote_node_folder_path on the parallel_decorator host if not existing
        with logger.info_context(f"Creating parallel_decorator mount point on {self.address}"):
            invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                         f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                         f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"mkdir -p {self.remote_project_mount_path}"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error creating parallel_decorator mount point on {self}.\n"
                                   f"Command: {repr(invocation)}. Ouput:\n{output}")

        # Mount the project root on remote_node_folder_path - use a separate process
        self.mount_project_remotely()

        with logger.info_context(f"Starting ray process on {self.address}"):
            invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                         f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                         f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"ray start --address {self.head_node.address}:{self.head_node.ray_port} " \
                         f"--ray-client-server-port {self.head_node.ray_port + 1} " \
                         f"--node-manager-port {self.head_node.ray_port + 2} " \
                         f"--object-manager-port {self.head_node.ray_port + 3} " \
                         f"--min-worker-port  {self.head_node.ray_port + 5} " \
                         f"--max-worker-port  {self.head_node.ray_port + self.head_node.ray_port_count - 1} " \
                         f"--redis-password='{self.head_node.session_password}' " \
                         + (f" --num-cpus {self.cpu_limit}" if self.cpu_limit else "")
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                raise RuntimeError(f"Error starting parallel_decorator ray on {self}.\n"
                                   f"Command: {repr(invocation)}. Ouput:\n{output}")

    def mount_project_remotely(self):
        # Mount the project root on remote_node_folder_path
        invocation = f"dpipe /usr/lib/openssh/sftp-server = " \
                     f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                     f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                     f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                     f"sshfs :{options.project_root} {self.remote_project_mount_path} -C -o sshfs_sync -o slave"

        self.mount_popen = subprocess.Popen(
            invocation, stdout=subprocess.PIPE,
            preexec_fn=os.setsid, shell=True)

        threading.Thread(target=self.mount_popen.communicate, daemon=True).start()

    def disconnect(self):
        assert not is_parallel_process()

        # Create remote_node_folder_path on the parallel_decorator host if not existing
        with logger.info_context(f"Disconnecting {self.address} (stopping ray)"):
            invocation = f"ssh -p {self.ssh_port if self.ssh_port else 22} " \
                         f"{'-i ' + self.local_ssh_file if self.local_ssh_file else ''} " \
                         f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                         f"ray stop --force"
            status, output = subprocess.getstatusoutput(invocation)
            if status != 0:
                logger.debug(f"Error disconnecting {self}. "
                             f"Command: {repr(invocation)}. Output:\n{output}")

        target_str = f"{self.ssh_user + '@' if self.ssh_user else ''}{self.address} " \
                     f"sshfs :{options.project_root}"
        for proc in psutil.process_iter():
            cmd_str = " ".join(proc.cmdline())
            if target_str in cmd_str:
                try:
                    os.kill(proc.pid, signal.SIGTERM)
                except ProcessLookupError:
                    logger.info(f"Cannot kill previously found process {proc.pid}")

    def __repr__(self):
        return f"{self.__class__.__name__}(address={self.address}, " \
               f"ssh_port={self.ssh_port}, ssh_user={self.ssh_user}, " \
               f"local_ssh_file={self.local_ssh_file}, " \
               f"cpu_limit={self.cpu_limit})"


def init_ray():
    """Initialize the ray cluster if it wasn't initialized before.
    """
    global _head_node

    if not ray.is_initialized():
        if _head_node is not None:
            _head_node.stop()

        # Initialize cluster of workers
        with logger.info_context(f"Initializing ray cluster [CPUlimit={options.ray_cpu_limit}]"):
            if not options.disable_swap:
                # From https://github.com/ray-project/ray/issues/10895 - allow using swap memory when needed,
                # avoiding early termination of jobs due to that.
                os.environ["RAY_DEBUG_DISABLE_MEMORY_MONITOR"] = "1"

            _head_node = HeadNode(ray_port=options.ray_port,
                                  ray_port_count=options.ray_port_count)
            _head_node.start()
            options.head_address = _head_node.get_node_ip()
            logger.verbose(_head_node.status_str)


def stop_ray():
    global _head_node
    if _head_node is not None:
        _head_node.stop()


def is_parallel_process():
    """Return True if and only if the call is made from a parallel_decorator ray process,
    which can be running in the head node or any of the parallel_decorator nodes (if any is present).
    """
    return is_ray_enabled() and os.path.basename(sys.argv[0]) == options.worker_script_name


def is_remote_node():
    """Return True if and only if the call is performed from a parallel_decorator ray process
    running on a node different from the head.
    """
    if not is_ray_enabled():
        return False
    if not is_parallel_process():
        print("Base process")
        return False
    else:
        try:
            return options._name_to_property["head_address"] != enb.misc.get_node_ip()
        except KeyError:
            return False


def is_ray_initialized():
    return is_ray_enabled() and ray.is_initialized


def parallel_decorator(*args, **kwargs):
    """Wrapper of the @`ray.remote` decorator that automatically updates enb.config.options
    for parallel_decorator processes, so that they always access the intended configuration.
    """
    kwargs["num_cpus"] = kwargs["num_cpus"] if "num_cpus" in kwargs else 1
    kwargs["num_gpus"] = kwargs["num_gpus"] if "num_gpus" in kwargs else 0

    def ray_remote_wrapper(f):
        enb.logger.debug(f"Wrapping {f} with ray")

        def remote_method_wrapper(_opts, *a, **k):
            """Wrapper for the decorated function f, that updates enb.config.options before f is called.
            """
            config.options.update(_opts, trigger_events=False)
            new_level = logger.get_level(logger.level_message.name, config.options.verbose)
            log.logger.selected_log_level = new_level
            log.logger.replace_print()

            return f(*a, **k)

        method_proxy = ray.remote(*args, **kwargs)(remote_method_wrapper)
        method_proxy.ray_remote = method_proxy.remote

        def local_side_remote(*a, **k):
            """Wrapper for ray's `.parallel_decorator()` method invoked in the local side.
            It makes sure that `remote_side_wrapper` receives the options argument.
            """
            try:
                try:
                    current_print = builtins.print
                    builtins.print = logger._original_print
                except AttributeError:
                    pass

                # apply ray.put to all arguments before passing them
                return method_proxy.ray_remote(
                    ray.put(dict(config.options.items())),
                    *[ray.put(argument) for argument in a],
                    **{key: ray.put(value) for key, value in k.items()})
            finally:
                builtins.print = current_print

        # method_proxy.remote = local_side_remote
        # method_proxy.start = method_proxy.remote
        del method_proxy.remote
        method_proxy.start = local_side_remote

        return method_proxy

    return ray_remote_wrapper


def get(ids, **kwargs):
    """Call ray's get method with the given arguments.
    """
    return ray.get(ids, **kwargs)


def get_completed_pending_ids(ids, timeout=0):
    """Return the list of completed and pending ids.
    """
    return ray.wait(ids, num_returns=len(ids), timeout=timeout)


def fix_imports():
    """An environment variable is passed to the children processes
    for them to be able to import all modules that were
    imported after loading enb. This prevents the remote
    functions to fail the deserialization process
    due to missing definitions.
    """
    if is_parallel_process():
        imported_modules = set()
        for module_name in sorted(ast.literal_eval(os.environ['_needed_modules'])):
            try:
                importlib.import_module(module_name)
                imported_modules.add(module_name)
            except ImportError as _ex:
                module_parts = module_name.split(".")
                for i in range(1, len(module_parts) - 1):
                    if ".".join(module_parts[:i]) in imported_modules:
                        break
                    else:
                        logger.error(f"Error importing module {repr(module_name)}: {repr(_ex)}.")



#!/usr/bin/env python3
"""Wrappers for the FAPEC codec.
"""
__author__ = "Miguel Hernández-Cabronero"
__since__ = "2020/05/25"

import os

import enb
from enb import icompression


class FAPEC_Abstract(icompression.LossyCodec, icompression.LosslessCodec, icompression.WrapperCodec):
    """Base class for FAPEC coders
    """
    BAND_FORMAT_BIP, BAND_FORMAT_BIL, BAND_FORMAT_BSQ, BAND_FORMAT_BAYER = range(4)
    default_band_format = BAND_FORMAT_BSQ

    def __init__(self, bin_dir=None, chunk_size_str="128M", threads=1,
                 lsb_discard_count=0,
                 adaptiveness_block_length=128,
                 output_invocation_dir=None):
        """
        :param bin_dir: path to the dir with the fapec and unfapec binaries
        :param chunk_size_str: string to be passed in the -chunk argument
        :param threads: number of threads to use for compression/decompression
        :param output_invocation_dir
        """
        param_dict = dict()
        param_dict["chunk"] = chunk_size_str
        assert lsb_discard_count >= 0
        if lsb_discard_count > 0:
            param_dict["lossy"] = lsb_discard_count
        assert threads >= 0
        param_dict["mt"] = threads
        assert 32 <= adaptiveness_block_length <= 1024
        param_dict["bl"] = adaptiveness_block_length

        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fapec"),
                         decompressor_path=os.path.join(bin_dir, "unfapec"),
                         param_dict=param_dict, output_invocation_dir=output_invocation_dir)

    def get_transform_dict_params(self, original_file_info):
        raise NotImplementedError("Please select one of the subclasses")

    def get_dtype(self, original_file_info):
        return 8 * original_file_info["bytes_per_sample"]

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        invocation_params = dict(self.param_dict)
        invocation_params.update(
            hd=0, chd=0,
            dtype=self.get_dtype(original_file_info=original_file_info))
        invocation_params.update(self.get_transform_dict_params(original_file_info=original_file_info))
        invocation = " ".join(f"-{k} {v}" for k, v in invocation_params.items())
        if original_file_info["signed"]:
            invocation += " -signed"
        if original_file_info["big_endian"]:
            invocation += " -be"
        invocation += " -ow -noattr "
        invocation += f"-o {compressed_path} {original_path}"
        return invocation

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        invocation = f"-o {reconstructed_path} -ow {compressed_path}"
        return invocation


class FAPEC_NP(FAPEC_Abstract):
    """Wrapper for FAPEC with no preprocessing (-np option).
    """

    def get_transform_dict_params(self, original_file_info):
        return dict(np=("tc" if original_file_info["signed"] else "us"))

    @property
    def label(self):
        return "FAPEC-NP"


class FAPEC_SPA(FAPEC_Abstract):
    """Wrapper for FAPEC with only spatial decorrelation (Delta)
    """

    def get_transform_dict_params(self, original_file_info):
        return dict()

    @property
    def label(self):
        return "FAPEC-SPA"


class FAPEC_MB(FAPEC_Abstract):
    """Wrapper for FAPEC with no preprocessing (-np option).
    """

    def get_transform_dict_params(self, original_file_info):
        return dict(bands=original_file_info["component_count"])

    @property
    def label(self):
        return "FAPEC-MB"


class FAPEC_LP(FAPEC_Abstract):
    """Wrapper for FAPEC with "spatial" lineal prediction (-lp flag)
    """

    def __init__(self, linear_prediction_order=1,
                 bin_dir=None, chunk_size_str="128M", threads=1,
                 lsb_discard_count=0,
                 adaptiveness_block_length=128,
                 output_invocation_dir=None):
        super().__init__(bin_dir=bin_dir, chunk_size_str=chunk_size_str,
                         threads=threads, lsb_discard_count=lsb_discard_count,
                         adaptiveness_block_length=adaptiveness_block_length,
                         output_invocation_dir=output_invocation_dir)
        assert linear_prediction_order >= 0
        assert linear_prediction_order == int(linear_prediction_order)
        self.linear_prediction_order = linear_prediction_order

    linear_prediction_oder = 1

    def get_transform_dict_params(self, original_file_info):
        assert self.linear_prediction_oder >= 1
        if self.linear_prediction_order == 1:
            return dict()
        else:
            return dict(od=self.linear_prediction_oder)

    @property
    def name(self):
        return super().name + (f"_od={self.linear_prediction_order}" if self.linear_prediction_order > 1 else "")

    @property
    def label(self):
        return "FAPEC-LP"


class FAPEC_DWT(FAPEC_Abstract):
    """Run FAPEC with the spectral+spatial IWT (-dwt flag).
    """

    def __init__(self, dwt_losses=0,
                 bin_dir=None, chunk_size_str="128M", threads=1,
                 lsb_discard_count=0,
                 adaptiveness_block_length=128,
                 output_invocation_dir=None):
        super().__init__(bin_dir=bin_dir, chunk_size_str=chunk_size_str,
                         threads=threads, lsb_discard_count=lsb_discard_count,
                         adaptiveness_block_length=adaptiveness_block_length,
                         output_invocation_dir=output_invocation_dir)
        self.dwt_losses = dwt_losses
        assert self.dwt_losses >= 0

    @property
    def name(self):
        return super().name + f"_dwt_losses={self.dwt_losses}"

    @property
    def label(self):
        return f"FAPEC-IWT"

    def get_transform_dict_params(self, original_file_info):
        return dict(dwt=f"{original_file_info['width']} "
                        f"{original_file_info['height']} "
                        f"{original_file_info['component_count']} "
                        f"{self.dwt_losses} "
                        f"{original_file_info['dynamic_range_bits']} "
                        f"{2}")  # <bfmt>    Bands format: 0=BIP, 1=BIL, 2=BSQ, 3=Bayer


class FAPEC_CILLIC(FAPEC_Abstract):
    """Run fapec with the CILLIC option.
    """

    def __init__(self, losses=0, multiband_levels=6,
                 bin_dir=None, chunk_size_str="128M", threads=1,
                 lsb_discard_count=0,
                 adaptiveness_block_length=128,
                 output_invocation_dir=None):
        super().__init__(bin_dir=bin_dir, chunk_size_str=chunk_size_str,
                         threads=threads, lsb_discard_count=lsb_discard_count,
                         adaptiveness_block_length=adaptiveness_block_length,
                         output_invocation_dir=output_invocation_dir)
        self.losses = losses
        self.multiband_levels = multiband_levels
        assert self.losses >= 0

    @property
    def name(self):
        return super().name + f"_losses={self.losses}_lev={self.multiband_levels}"

    @property
    def label(self):
        return f"FAPEC-CILLIC"

    def get_transform_dict_params(self, original_file_info):
        return dict(cillic=f"{original_file_info['width']} "
                           f"{original_file_info['height']} "
                           f"{original_file_info['component_count']} "
                           f"{self.losses} "
                           f"{original_file_info['dynamic_range_bits']} "
                           f"{2}",  # <bfmt>    Bands format: 0=BIP, 1=BIL, 2=BSQ, 3=Bayer
                    lev=self.multiband_levels)


class FAPEC_2DDWT(FAPEC_DWT):
    """Run FAPEC with a single spatial DWT transform, by forcing dymensions
    to be z'=1, x'=x, y'=y*z.
    """

    @property
    def name(self):
        return super().name + f"_dwt_losses={self.dwt_losses}"

    @property
    def label(self):
        return f"FAPEC-2DIWT"

    def get_transform_dict_params(self, original_file_info):

        dimension_str = f"{original_file_info['width']} " \
                        f"{original_file_info['height'] * original_file_info['component_count']} " \
                        f"1 "

        if original_file_info["component_count"] < 8461:
            return dict(dwt=f"{dimension_str} "
                            f"{self.dwt_losses} "
                            f"{original_file_info['dynamic_range_bits']} "
                            f"{2}")  # <bfmt>    Bands format: 0=BIP, 1=BIL, 2=BSQ, 3=Bayer
        else:
            # IASI crashes otherwise
            return dict()


class FAPEC_HPA(FAPEC_Abstract):
    def __init__(self, hpa_losses=0,
                 bin_dir=None, chunk_size_str="128M", threads=1,
                 lsb_discard_count=0,
                 adaptiveness_block_length=128,
                 output_invocation_dir=None):
        super().__init__(bin_dir=bin_dir, chunk_size_str=chunk_size_str,
                         threads=threads, lsb_discard_count=lsb_discard_count,
                         adaptiveness_block_length=adaptiveness_block_length,
                         output_invocation_dir=output_invocation_dir)
        self.hpa_losses = hpa_losses
        assert 0 <= self.hpa_losses <= 16

    @property
    def name(self):
        return super().name + f"_hpa_losses={self.hpa_losses}"

    @property
    def label(self):
        return f"FAPEC-HPA"

    def get_transform_dict_params(self, original_file_info):
        return dict(hpa=f"{original_file_info['width']} "
                        f"{original_file_info['height']} "
                        f"{original_file_info['component_count']} "
                        f"{self.hpa_losses} "
                        f"{original_file_info['dynamic_range_bits']} "
                        f"{2}")  # <bfmt>    Bands format: 0=BIP, 1=BIL, 2=BSQ, 3=Bayer


class FAPEC_FITS(enb.icompression.LosslessCodec, enb.icompression.FITSWrapperCodec):
    def __init__(self,
                 bin_dir=None):
        bin_dir = bin_dir if bin_dir is not None else os.path.dirname(__file__)
        super().__init__(compressor_path=os.path.join(bin_dir, "fapec"),
                         decompressor_path=os.path.join(bin_dir, "unfapec"),
                         param_dict=dict())

    @property
    def name(self):
        """Don't include the binary signature
        """
        name = f"{self.__class__.__name__}"
        return name

    @property
    def label(self):
        return "FAPEC-FITS"

    def get_compression_params(self, original_path, compressed_path, original_file_info):
        return f"-o {compressed_path} -ow {original_path}  "

    def get_decompression_params(self, compressed_path, reconstructed_path, original_file_info):
        return f" -o {reconstructed_path} -ow {compressed_path} "
